from abc import abstractmethod


class activity_model(object):
    """ The activity model object is for implementing activity models
    for volatile species in melts. It contains all the methods required to
    evaluate the activity.
    """
    def __init__(self):
        self.set_calibration_ranges([])

    def set_calibration_ranges(self, calibration_ranges):
        self.calibration_ranges = calibration_ranges

    @abstractmethod
    def activity(self, X, **kwargs):
        """
        """

    # @abstractmethod
    def check_calibration_range(self, parameters, report_nonexistance=True):
        s = ''
        for cr in self.calibration_ranges:
            if cr.check(parameters) is False:
                s += cr.string(parameters, report_nonexistance)
        return s


# --------------- ACTVITY MODELS ------------------------------- #

class activity_idealsolution(activity_model):
    """ Implements an ideal solution activity model, i.e. it
    will always return the mole fraction.
    """

    def activity(self, X):
        """ The activity of the component in an ideal solution, i.e., it
        will return the mole fraction.

        Parameters
        ----------
        X   float
            The mole fraction of the species in the solution.

        Returns
        -------
        float
            The activity of the species in the solution, i.e.,
            the mole fraction.
        """
        return X



from VESIcal import calibration_checks
from VESIcal import core

from scipy.optimize import root_scalar
from abc import abstractmethod
import numpy as np


class FugacityModel(object):
    """ The fugacity model object is for implementations of fugacity models
    for individual volatile species, though it may depend on the mole
    fraction of other volatile species. It contains all the methods required
    to calculate the fugacity at a given pressure and mole fraction.
    """

    def __init__(self):
        self.set_calibration_ranges([])

    def set_calibration_ranges(self, calibration_ranges):
        self.calibration_ranges = calibration_ranges

    @abstractmethod
    def fugacity(self, pressure, **kwargs):
        """
        """

    # @abstractmethod
    def check_calibration_range(self, parameters, report_nonexistance=True):
        s = ''
        for cr in self.calibration_ranges:
            if cr.check(parameters) is False:
                s += cr.string(parameters, report_nonexistance)
        return s


# ------------- FUGACITY MODELS -------------------------------- #

class fugacity_idealgas(FugacityModel):
    """ An instance of FugacityModel for an ideal gas.
    """

    def fugacity(self, pressure, X_fluid=1.0, **kwargs):
        """ Returns the fugacity of an ideal gas, i.e., the partial pressure.

        Parameters
        ----------
        pressure    float
            Total pressure of the system, in bars.
        X_fluid     float
            The mole fraction of the species in the vapour phase.

        Returns
        -------
        float
            Fugacity (partial pressure) in bars
        """
        return pressure*X_fluid


class fugacity_KJ81_co2(FugacityModel):
    """ Implementation of the Kerrick and Jacobs (1981) EOS for mixed fluids. This class
    will return the properties of the CO2 component of the mixed fluid.
    """
    def __init__(self):
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', 20000.0,  calibration_checks.crf_LessThan, 'bar',
                'Kerrick and Jacobs (1981) EOS', fail_msg=calibration_checks.crmsg_LessThan_fail,
                pass_msg=calibration_checks.crmsg_LessThan_pass,
                description_msg=calibration_checks.crmsg_LessThan_description),
            calibration_checks.CalibrationRange(
                'temperature', 1050, calibration_checks.crf_LessThan, 'oC',
                'Kerrick and Jacobs (1981) EOS', fail_msg=calibration_checks.crmsg_LessThan_fail,
                pass_msg=calibration_checks.crmsg_LessThan_pass,
                description_msg=calibration_checks.crmsg_LessThan_description)])

    def fugacity(self, pressure, temperature, X_fluid, **kwargs):
        """ Calculates the fugacity of CO2 in a mixed CO2-H2O fluid. Above 1050C, it assumes H2O
        and CO2 do not interact, as the equations are not defined beyond this point.

        Parameters
        ----------
        pressure    float
            Total pressure of the system in bars.
        temperature     float
            Temperature in degC
        X_fluid     float
            Mole fraction of CO2 in the fluid.

        Returns
        -------
        float
            fugacity of CO2 in bars
        """
        if X_fluid == 0:
            return 0
        elif temperature >= 1050.0:
            return pressure*np.exp(self.lnPhi_mix(pressure, temperature, 1.0))*X_fluid
        else:
            return pressure*np.exp(self.lnPhi_mix(pressure, temperature, X_fluid))*X_fluid

    def volume(self, P, T, X_fluid):
        """ Calculates the volume of the mixed fluid, by solving Eq (28) of Kerrick and
        Jacobs (1981) using scipy.root_scalar.

        Parameters
        ----------
        P   float
            Total pressure of the system, in bars.
        T   float
            Temperature in degC
        X_fluid     float
            Mole fraction of CO2 in the fluid

        Returns
        -------
        float
            Volume of the mixed fluid.
        """
        if X_fluid != 1.0:
            # x0 = self.volume(P,T,1.0)*X_fluid + self.volume_h(P,T)*(1-X_fluid)
            # print(x0)
            if P >= 20000 and T < 800-273.15:
                x0 = (X_fluid*25+(1-X_fluid)*15)
            else:
                x0 = (X_fluid*35+(1-X_fluid)*15)

        else:
            if P >= 20000 and T < 800-273.15:
                x0 = 25
            else:
                x0 = 35
        return root_scalar(self.root_volume, x0=x0, x1=x0*0.9, args=(P, T, X_fluid)).root

    def root_volume(self, v, P, T, X_fluid):
        """ Returns the difference between the lhs and rhs of Eq (28) of Kerrick and Jacobs (1981).
        For use with a root finder to obtain the volume of the mixed fluid.

        Parameters
        ----------
        v   float
            Guess for the volume
        P   float
            Total system pressure in bars.
        T   float
            Temperature in degC
        X_fluid     float
            Mole fraction of CO2 in the fluid.

        Returns
        -------
        float
            Difference between lhs and rhs of Eq (28) of Kerrick and Jacobs (1981), in bars.
        """
        T = T + 273.15
        c = {}
        h = {}

        c['b'] = 58.0
        c['c'] = (28.31 + 0.10721*T - 8.81e-6*T**2)*1e6
        c['d'] = (9380.0 - 8.53*T + 1.189e-3*T**2)*1e6
        c['e'] = (-368654.0 + 715.9*T + 0.1534*T**2)*1e6
        h['b'] = 29.0
        h['c'] = (290.78 - 0.30276*T + 1.4774e-4*T**2)*1e6
        h['d'] = (-8374.0 + 19.437*T - 8.148e-3*T**2)*1e6
        h['e'] = (76600.0 - 133.9*T + 0.1071*T**2)*1e6

        if X_fluid == 1:
            bm = c['b']
            cm = c['c']
            c12 = c['c']
            dm = c['d']
            d12 = c['d']
            em = c['e']
            e12 = c['e']
        else:
            bm = X_fluid*c['b'] + (1-X_fluid)*h['b']
            c12 = (c['c']*h['c'])**0.5
            cm = c['c']*X_fluid**2 + h['c']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*c12
            d12 = (c['d']*h['d'])**0.5
            dm = c['d']*X_fluid**2 + h['d']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*d12
            e12 = (c['e']*h['e'])**0.5
            em = c['e']*X_fluid**2 + h['e']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*e12

        am = cm + dm/v + em/v**2

        y = bm/(4*v)

        pt1 = (83.14 * T * (1 + y + y**2 - y**3)) / (v*(1-y)**3)
        pt2 = - am / (T**0.5 * v * (v+bm))

        return -(P - pt1 - pt2)

    def volume_h(self, P, T):
        """ Calculates the volume of a pure H2O fluid, by solving Eq (14) of
        Kerrick and Jacobs (1981).

        Parameters
        ----------
        P   float
            Total pressure in bars.
        T   float
            Temperature in degC.

        Returns
        -------
        Difference between lhs and rhs of Eq (14) of Kerrick and Jacobs (1981), in bars.
        """
        return root_scalar(self.root_volume_h, x0=15, x1=35, args=(P, T)).root

    def root_volume_h(self, v, P, T):
        """ Returns the difference between the lhs and rhs of Eq (14) of
        Kerrick and Jacobs (1981). For use with a root solver to identify the
        volume of a pure H2O fluid.

        Parameters
        ----------
        v   float
            Guess for the volume
        P   float
            Total pressure in bars.
        T   float
            Temperature in degC.

        Returns
        -------
        float
            The difference between the lhs and rhs of Eq (14) of Kerrick and Jacobs (1981),
            in bars.
        """
        T = T + 273.15
        h = {}
        h['b'] = 29.0
        h['c'] = (290.78 - 0.30276*T + 1.4774e-4*T**2)*1e6
        h['d'] = (-8374.0 + 19.437*T - 8.148e-3*T**2)*1e6
        h['e'] = (76600.0 - 133.9*T + 0.1071*T**2)*1e6
        h['a'] = h['c'] + h['d']/v + h['e']/v**2

        y = h['b']/(4*v)

        pt1 = (83.14 * T * (1 + y + y**2 - y**3)) / (v*(1-y)**3)
        pt2 = - h['a'] / (T**0.5 * v * (v+h['b']))

        return -(P - pt1 - pt2)

    def lnPhi_mix(self, P, T, X_fluid):
        """ Calculates the natural log of the fugacity coefficient for CO2 in a
        mixed CO2-H2O fluid. Uses Eq (27) of Kerrick and Jacobs (1981).

        Parameters
        ----------
        P   float
            Total pressure in bars.
        T   float
            Temperature in degC
        X_fluid     float
            The mole fraction of CO2 in the fluid.

        Returns
        -------
        float
            The natural log of the fugacity coefficient for CO2 in a mixed fluid.
        """
        T = T + 273.15
        v = self.volume(P, T-273.15, X_fluid)

        c = {}
        h = {}

        c['b'] = 58.0
        c['c'] = (28.31 + 0.10721*T - 8.81e-6*T**2)*1e6
        c['d'] = (9380.0 - 8.53*T + 1.189e-3*T**2)*1e6
        c['e'] = (-368654.0 + 715.9*T + 0.1534*T**2)*1e6
        h['b'] = 29.0
        h['c'] = (290.78 - 0.30276*T + 1.4774e-4*T**2)*1e6
        h['d'] = (-8374.0 + 19.437*T - 8.148e-3*T**2)*1e6
        h['e'] = (76600.0 - 133.9*T + 0.1071*T**2)*1e6

        if X_fluid == 1:
            bm = c['b']
            cm = c['c']
            c12 = c['c']
            dm = c['d']
            d12 = c['d']
            em = c['e']
            e12 = c['e']
        else:
            bm = X_fluid*c['b'] + (1-X_fluid)*h['b']
            c12 = (c['c']*h['c'])**0.5
            cm = c['c']*X_fluid**2 + h['c']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*c12
            d12 = (c['d']*h['d'])**0.5
            dm = c['d']*X_fluid**2 + h['d']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*d12
            e12 = (c['e']*h['e'])**0.5
            em = c['e']*X_fluid**2 + h['e']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*e12

        y = bm/(4*v)

        Z = v*P/(83.14*T)

        lnPhi = 0

        lnPhi += (4*y-3*y**2)/(1-y)**2 + (c['b']/bm * (4*y-2*y**2)/(1-y)**3)
        lnPhi += - (2*c['c']*X_fluid+2*(1-X_fluid)*c12)/(83.14*T**1.5*bm)*np.log((v+bm)/v)
        lnPhi += - cm*c['b']/(83.14*T**1.5*bm*(v+bm))
        lnPhi += cm*c['b']/(83.14*T**1.5*bm**2)*np.log((v+bm)/v)
        lnPhi += - (2*c['d']*X_fluid+2*d12*(1-X_fluid)+dm)/(83.14*T**1.5*bm*v)
        lnPhi += (2*c['d']*X_fluid+2*(1-X_fluid)*d12+dm)/(83.14*T**1.5*bm**2)*np.log((v+bm)/v)
        lnPhi += c['b']*dm/(83.14*T**1.5*v*bm*(v+bm)) + 2*c['b']*dm/(83.14*T**1.5*bm**2*(v+bm))
        lnPhi += - 2*c['b']*dm/(83.14*T**1.5*bm**3)*np.log((v+bm)/v)
        lnPhi += - (2*c['e']*X_fluid + 2*(1-X_fluid)*e12+2*em)/(83.14*T**1.5*2*bm*v**2)
        lnPhi += (2*c['e']*X_fluid+2*e12*(1-X_fluid)+2*em)/(83.14*T**1.5*bm**2*v)
        lnPhi += - (2*c['e']*X_fluid+2*e12*(1-X_fluid)+2*em)/(83.14*T**1.5*bm**3)*np.log((v+bm)/v)
        lnPhi += (em*c['b']/(83.14*T**1.5*2*bm*v**2*(v+bm)) -
                  3*em*c['b']/(83.14*T**1.5*2*bm**2*v*(v+bm)))
        lnPhi += (3*em*c['b']/(83.14*T**1.5*bm**4)*np.log((v+bm)/v) -
                  3*em*c['b']/(83.14*T**1.5*bm**3*(v+bm)))
        lnPhi += - np.log(Z)

        return lnPhi


class fugacity_KJ81_h2o(FugacityModel):
    """Implementation of the Kerrick and Jacobs (1981) EOS for mixed fluids. This class
    will return the properties of the H2O component of the mixed fluid.
    """
    def __init__(self):
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', 20000.0, calibration_checks.crf_LessThan, 'bar',
                'Kerrick and Jacobs (1981) EOS',
                fail_msg=calibration_checks.crmsg_LessThan_fail,
                pass_msg=calibration_checks.crmsg_LessThan_pass,
                description_msg=calibration_checks.crmsg_LessThan_description),
            calibration_checks.CalibrationRange(
                'temperature', 1050, calibration_checks.crf_LessThan, 'oC',
                'Kerrick and Jacobs (1981) EOS',
                fail_msg=calibration_checks.crmsg_LessThan_fail,
                pass_msg=calibration_checks.crmsg_LessThan_pass,
                description_msg=calibration_checks.crmsg_LessThan_description)])

    def fugacity(self, pressure, temperature, X_fluid, **kwargs):
        """ Calculates the fugacity of H2O in a mixed CO2-H2O fluid. Above 1050C,
        it assumes H2O and CO2 do not interact, as the equations are not defined
        beyond this point.

        Parameters
        ----------
        pressure    float
            Total pressure of the system in bars.
        temperature     float
            Temperature in degC
        X_fluid     float
            Mole fraction of H2O in the fluid.

        Returns
        -------
        float
            fugacity of H2O in bars
        """
        if X_fluid == 0:
            return 0
        elif temperature >= 1050:
            return pressure*np.exp(self.lnPhi_mix(pressure, temperature, 1.0))*X_fluid
        else:
            return pressure*np.exp(self.lnPhi_mix(pressure, temperature, X_fluid))*X_fluid

    def volume(self, P, T, X_fluid):
        """ Calculates the volume of the mixed fluid, by solving Eq (28) of Kerrick and
        Jacobs (1981) using scipy.root_scalar.

        Parameters
        ----------
        P   float
            Total pressure of the system, in bars.
        T   float
            Temperature in degC
        X_fluid     float
            Mole fraction of H2O in the fluid

        Returns
        -------
        float
            Volume of the mixed fluid.
        """
        if X_fluid != 1.0:
            if P >= 20000 and T < 800-273.15:
                x0 = ((1-X_fluid)*25+X_fluid*15)
            else:
                x0 = ((1-X_fluid)*35+X_fluid*15)

        else:
            if P >= 20000 and T < 800-273.15:
                x0 = 10
            else:
                x0 = 15
        return root_scalar(self.root_volume, x0=x0, x1=x0*0.9, args=(P, T, X_fluid)).root

    def root_volume(self, v, P, T, X_fluid):
        """ Returns the difference between the lhs and rhs of Eq (28) of Kerrick and Jacobs (1981).
        For use with a root finder to obtain the volume of the mixed fluid.

        Parameters
        ----------
        v   float
            Guess for the volume
        P   float
            Total system pressure in bars.
        T   float
            Temperature in degC
        X_fluid     float
            Mole fraction of H2O in the fluid.

        Returns
        -------
        float
            Difference between lhs and rhs of Eq (28) of Kerrick and Jacobs (1981), in bars.
        """
        T = T + 273.15
        c = {}
        h = {}

        c['b'] = 58.0
        c['c'] = (28.31 + 0.10721*T - 8.81e-6*T**2)*1e6
        c['d'] = (9380.0 - 8.53*T + 1.189e-3*T**2)*1e6
        c['e'] = (-368654.0 + 715.9*T + 0.1534*T**2)*1e6
        h['b'] = 29.0
        h['c'] = (290.78 - 0.30276*T + 1.4774e-4*T**2)*1e6
        h['d'] = (-8374.0 + 19.437*T - 8.148e-3*T**2)*1e6
        h['e'] = (76600.0 - 133.9*T + 0.1071*T**2)*1e6

        if X_fluid == 1:
            bm = h['b']
            cm = h['c']
            dm = h['d']
            em = h['e']
            c12 = h['c']
            d12 = h['d']
            e12 = h['e']
        else:
            bm = X_fluid*h['b'] + (1-X_fluid)*c['b']
            c12 = (c['c']*h['c'])**0.5
            cm = h['c']*X_fluid**2 + c['c']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*c12
            d12 = (c['d']*h['d'])**0.5
            dm = h['d']*X_fluid**2 + c['d']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*d12
            e12 = (c['e']*h['e'])**0.5
            em = h['e']*X_fluid**2 + c['e']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*e12

        am = cm + dm/v + em/v**2

        y = bm/(4*v)

        pt1 = (83.14 * T * (1 + y + y**2 - y**3)) / (v*(1-y)**3)
        pt2 = - am / (T**0.5 * v * (v+bm))

        return -(P - pt1 - pt2)

    def volume_c(self, P, T):
        """ Calculates the volume of a pure CO2 fluid, by solving Eq (14) of
        Kerrick and Jacobs (1981).

        Parameters
        ----------
        P   float
            Total pressure in bars.
        T   float
            Temperature in degC.

        Returns
        -------
        Difference between lhs and rhs of Eq (14) of Kerrick and Jacobs (1981), in bars.
        """
        return root_scalar(self.root_volume_c, x0=15, x1=35, args=(P, T)).root

    def root_volume_c(self, v, P, T):
        """ Returns the difference between the lhs and rhs of Eq (14) of
        Kerrick and Jacobs (1981). For use with a root solver to identify the
        volume of a pure H2O fluid.

        Parameters
        ----------
        v   float
            Guess for the volume
        P   float
            Total pressure in bars.
        T   float
            Temperature in degC.

        Returns
        -------
        float
            The difference between the lhs and rhs of Eq (14) of Kerrick and Jacobs (1981),
            in bars.
        """
        T = T + 273.15
        c = {}
        c['b'] = 58.0
        c['c'] = (28.31 + 0.10721*T - 8.81e-6*T**2)*1e6
        c['d'] = (9380.0 - 8.53*T + 1.189e-3*T**2)*1e6
        c['e'] = (-368654.0 + 715.9*T + 0.1534*T**2)*1e6
        c['a'] = c['c'] + c['d']/v + c['e']/v**2

        y = c['b']/(4*v)

        pt1 = (83.14 * T * (1 + y + y**2 - y**3)) / (v*(1-y)**3)
        pt2 = - c['a'] / (T**0.5 * v * (v+c['b']))

        return -(P - pt1 - pt2)

    def lnPhi_mix(self, P, T, X_fluid):
        """ Calculates the natural log of the fugacity coefficient for H2O in a
        mixed CO2-H2O fluid. Uses Eq (27) of Kerrick and Jacobs (1981).

        Parameters
        ----------
        P   float
            Total pressure in bars.
        T   float
            Temperature in degC
        X_fluid     float
            The mole fraction of H2O in the fluid.

        Returns
        -------
        float
            The natural log of the fugacity coefficient for H2O in a mixed fluid.
        """
        T = T + 273.15
        v = self.volume(P, T-273.15, X_fluid)

        c = {}
        h = {}

        c['b'] = 58.0
        c['c'] = (28.31 + 0.10721*T - 8.81e-6*T**2)*1e6
        c['d'] = (9380.0 - 8.53*T + 1.189e-3*T**2)*1e6
        c['e'] = (-368654.0 + 715.9*T + 0.1534*T**2)*1e6
        h['b'] = 29.0
        h['c'] = (290.78 - 0.30276*T + 1.4774e-4*T**2)*1e6
        h['d'] = (-8374.0 + 19.437*T - 8.148e-3*T**2)*1e6
        h['e'] = (76600.0 - 133.9*T + 0.1071*T**2)*1e6

        if X_fluid == 1:
            bm = h['b']
            cm = h['c']
            dm = h['d']
            em = h['e']
            c12 = h['c']
            d12 = h['d']
            e12 = h['e']
        else:
            bm = X_fluid*h['b'] + (1-X_fluid)*c['b']
            c12 = (c['c']*h['c'])**0.5
            cm = h['c']*X_fluid**2 + c['c']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*c12
            d12 = (c['d']*h['d'])**0.5
            dm = h['d']*X_fluid**2 + c['d']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*d12
            e12 = (c['e']*h['e'])**0.5
            em = h['e']*X_fluid**2 + c['e']*(1-X_fluid)**2 + 2*X_fluid*(1-X_fluid)*e12

        y = bm/(4*v)

        # Z = (1+y+y**2-y**3)/(1-y)**2 - am/(83.14*T**1.5*(v+bm))
        Z = v*P/(83.14*T)

        lnPhi = 0

        lnPhi += (4*y-3*y**2)/(1-y)**2 + (h['b']/bm * (4*y-2*y**2)/(1-y)**3)
        lnPhi += - (2*h['c']*X_fluid+2*(1-X_fluid)*c12)/(83.14*T**1.5*bm)*np.log((v+bm)/v)
        lnPhi += - cm*h['b']/(83.14*T**1.5*bm*(v+bm))
        lnPhi += cm*h['b']/(83.14*T**1.5*bm**2)*np.log((v+bm)/v)
        lnPhi += - (2*h['d']*X_fluid+2*d12*(1-X_fluid)+dm)/(83.14*T**1.5*bm*v)
        lnPhi += (2*h['d']*X_fluid+2*(1-X_fluid)*d12+dm)/(83.14*T**1.5*bm**2)*np.log((v+bm)/v)
        lnPhi += h['b']*dm/(83.14*T**1.5*v*bm*(v+bm)) + 2*h['b']*dm/(83.14*T**1.5*bm**2*(v+bm))
        lnPhi += - 2*h['b']*dm/(83.14*T**1.5*bm**3)*np.log((v+bm)/v)
        lnPhi += - (2*h['e']*X_fluid + 2*(1-X_fluid)*e12+2*em)/(83.14*T**1.5*2*bm*v**2)
        lnPhi += (2*h['e']*X_fluid+2*e12*(1-X_fluid)+2*em)/(83.14*T**1.5*bm**2*v)
        lnPhi += - (2*h['e']*X_fluid+2*e12*(1-X_fluid)+2*em)/(83.14*T**1.5*bm**3)*np.log((v+bm)/v)
        lnPhi += (em*h['b']/(83.14*T**1.5*2*bm*v**2*(v+bm)) -
                  3*em*h['b']/(83.14*T**1.5*2*bm**2*v*(v+bm)))
        lnPhi += (3*em*h['b']/(83.14*T**1.5*bm**4)*np.log((v+bm)/v) -
                  3*em*h['b']/(83.14*T**1.5*bm**3*(v+bm)))
        lnPhi += - np.log(Z)

        return lnPhi


class fugacity_ZD09_co2(FugacityModel):
    """ Implementation of the Zhang and Duan (2009) fugacity model for pure CO2
    fluids."""
    def __init__(self):
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', [1, 1e5], calibration_checks.crf_Between, 'bar',
                'Zhang and Duan (2009) EOS',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'temperature', [200, 2300], calibration_checks.crf_Between, 'oC',
                'Zhang and Duan (2009) EOS',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description)])

    def fugacity(self, pressure, temperature, X_fluid=1.0, **kwargs):
        """ Calculates the fugacity of a pure CO2 fluid, or a mixed fluid assuming
        ideal mixing. Implements eqn (14) of Zhang and Duan (2009).

        Parameters
        ---------
        pressure     float
            Pressure in bars
        temperature     float
            Temperature in degC
        X_fluid     float
            Mole fraction of CO2 in the fluid. Default is 1.0.

        Returns
        -------
        float
            Fugacity of CO2, standard state 1 bar.
        """

        P = pressure/10
        T = temperature + 273.15

        a = np.array([0.0,
                      2.95177298930e-2,
                      -6.33756452413e3,
                      -2.75265428882e5,
                      1.29128089283e-3,
                      -1.45797416153e2,
                      7.65938947237e4,
                      2.58661493537e-6,
                      0.52126532146,
                      -1.39839523753e2,
                      -2.36335007175e-8,
                      5.35026383543e-3,
                      -0.27110649951,
                      2.50387836486e4,
                      0.73226726041,
                      1.5483335997e-2])
        e = 235.0
        s = 3.79

        Pm = 3.0636*P*s**3/e
        Tm = 154*T/e
        Vm = root_scalar(self.Vm, x0=200, x1=100, args=(P, T)).root

        S1 = ((a[1]+a[2]/Tm**2+a[3]/Tm**3)/Vm +
              (a[4]+a[5]/Tm**2+a[6]/Tm**3)/(2*Vm**2) +
              (a[7]+a[8]/Tm**2+a[9]/Tm**3)/(4*Vm**4) +
              (a[10]+a[11]/Tm**2+a[12]/Tm**3)/(5*Vm**5) +
              (a[13]/(2*a[15]*Tm**3)*(a[14]+1-(a[14]+1+a[15]/Vm**2) *
               np.exp(-a[15]/Vm**2)))
              )

        Z = Pm*Vm/(8.314*Tm)

        lnfc = Z - 1 - np.log(Z) + S1

        return P*np.exp(lnfc)*10

    def Vm(self, Vm, P, T):
        """ Function to use for solving for the parameter Vm, defined by eqn (8) of
        Zhang and Duan (2009). Called by scipy.fsolve in the fugacity method.

        Parameters
        ----------
        Vm     float
            Guessed value of Vm
        P     float
            Pressure in MPa
        T     float
            Temperature in K

        Returns
        -------
        float
            Difference between (rearranged) LHS and RHS of eqn (8) of Zhang and Duan (2009).
        """
        Pm = 3.0636*P*3.79**3/235.0
        Tm = 154*T/235.0
        a = np.array([0.0,
                      2.95177298930e-2,
                      -6.33756452413e3,
                      -2.75265428882e5,
                      1.29128089283e-3,
                      -1.45797416153e2,
                      7.65938947237e4,
                      2.58661493537e-6,
                      0.52126532146,
                      -1.39839523753e2,
                      -2.36335007175e-8,
                      5.35026383543e-3,
                      -0.27110649951,
                      2.50387836486e4,
                      0.73226726041,
                      1.5483335997e-2])

        return ((1+(a[1]+a[2]/Tm**2+a[3]/Tm**3)/Vm +
                 (a[4]+a[5]/Tm**2+a[6]/Tm**3)/Vm**2 +
                 (a[7]+a[8]/Tm**2+a[9]/Tm**3)/Vm**4)*0.08314*Tm/Pm - Vm
                )


class fugacity_MRK_co2(FugacityModel):
    """ Modified Redlick Kwong fugacity model as used by VolatileCalc. Python implementation by
    D. J. Rasmussen (github.com/DJRgeoscience/VolatileCalcForPython), based on VB code by Newman &
    Lowenstern.
    """
    def __init__(self):
        self.set_calibration_ranges([])

    def fugacity(self, pressure, temperature, X_fluid=1.0, **kwargs):
        """ Calculates the fugacity of CO2 in a pure or mixed H2O-CO2 fluid (assuming ideal
        mixing).

        Parameters
        ----------
        pressure    float
            Total pressure of the system in bars.
        temperature     float
            Temperature in degC
        X_fluid     float
            Mole fraction of CO2 in the fluid.

        Returns
        -------
        float
            fugacity of CO2 in bars
        """
        fug = self.MRK(pressure, temperature+273.15)
        return fug*X_fluid

    def FNA(self, TK):
        return ((166800000 - 193080 * (TK - 273.15) + 186.4 * (TK - 273.15)**2
                - 0.071288 * ((TK - 273.15)**3)) * 1.01325)

    def FNB(self, TK):
        return 1.01325 * (73030000 - 71400 * (TK - 273.15) + 21.57 * (TK - 273.15)**2)

    def FNC(self, TK):
        R = 83.14321
        return (1.01325 * (np.exp(-11.071 + 5953 / TK - 2746000 / TK**2 + 464600000 / TK**3)
                * 0.5 * R * R * TK**2.5 / 1.02668 + 40123800))

    def FNF(self, V, TK, A, B, P):
        R = 83.14321
        return R * TK / (V - B) - A / ((V * V + B * V) * TK**0.5) - P

    def MRK(self, P, TK):  # Redlich-Kwong routine to estimate endmember H2O and CO2 fugacities
        R = 83.14321
        B_1 = 14.6
        B_2 = 29.7

        for X_1 in [0, 1]:
            B = X_1 * B_1 + (1 - X_1) * B_2
            A = (X_1**2 * self.FNA(TK) + 2 * X_1 * (1 - X_1) * self.FNC(TK) +
                 (1 - X_1)**2 * self.FNB(TK))
            Temp2 = B + 5
            Q = 1
            Temp1 = 0
            while abs(Temp2 - Temp1) >= 0.00001:
                Temp1 = Temp2
                F_1 = (self.FNF(Temp1 + 0.01, TK, A, B, P) - self.FNF(Temp1, TK, A, B, P)) / 0.01
                Temp2 = Temp1 - Q * self.FNF(Temp1, TK, A, B, P) / F_1
                F_2 = (self.FNF(Temp2 + 0.01, TK, A, B, P) - self.FNF(Temp2, TK, A, B, P)) / 0.01
                if F_2 * F_1 <= 0:
                    Q = Q / 2.
                if abs(Temp2 - Temp1) > 0.00001:
                    F_1 = F_2
            V = Temp2
            G_1 = (np.log(V / (V - B)) + B_1 / (V - B) - 2 * (X_1 * self.FNA(TK) +
                   (1 - X_1) * self.FNC(TK)) * np.log((V + B) / V) / (R * TK**1.5 * B))
            G_1 = (G_1 + (np.log((V + B) / V) - B / (V + B)) * A * B_1 / (R * TK**1.5 * B**2) -
                   np.log(P * V / (R * TK)))
            G_1 = np.exp(G_1)
            G_2 = (np.log(V / (V - B)) + B_2 / (V - B) - 2 * (X_1 * self.FNC(TK) +
                   (1 - X_1) * self.FNB(TK)) * np.log((V + B) / V) / (R * TK**1.5 * B))
            G_2 = (G_2 + (np.log((V + B) / V) - B / (V + B)) * A * B_2 / (R * TK**1.5 * B**2) -
                   np.log(P * V / (R * TK)))
            G_2 = np.exp(G_2)
            if X_1 == 0:
                fCO2o = G_2 * P  # The fugacity of CO2
        # return fCO2o
        return fCO2o


class fugacity_MRK_h2o(FugacityModel):
    """ Modified Redlick Kwong fugacity model as used by VolatileCalc. Python implementation by
    D. J. Rasmussen (github.com/DJRgeoscience/VolatileCalcForPython), based on VB code by Newman &
    Lowenstern.
    """
    def __init__(self):
        self.set_calibration_ranges([])

    def fugacity(self, pressure, temperature, X_fluid=1.0, **kwargs):
        """ Calculates the fugacity of H2O in a pure or mixed H2O-CO2 fluid (assuming ideal
        mixing).

        Parameters
        ----------
        pressure    float
            Total pressure of the system in bars.
        temperature     float
            Temperature in degC
        X_fluid     float
            Mole fraction of H2O in the fluid.

        Returns
        -------
        float
            fugacity of H2O in bars
        """
        fug = self.MRK(pressure, temperature+273.15)
        return fug*X_fluid

    def FNA(self, TK):
        return ((166800000 - 193080 * (TK - 273.15) + 186.4 * (TK - 273.15)**2 -
                0.071288 * ((TK - 273.15)**3)) * 1.01325)

    def FNB(self, TK):
        return 1.01325 * (73030000 - 71400 * (TK - 273.15) + 21.57 * (TK - 273.15)**2)

    def FNC(self, TK):
        R = 83.14321
        return (1.01325 * (np.exp(-11.071 + 5953 / TK - 2746000 / TK**2 + 464600000 / TK**3) *
                0.5 * R * R * TK**2.5 / 1.02668 + 40123800))

    def FNF(self, V, TK, A, B, P):
        R = 83.14321
        return R * TK / (V - B) - A / ((V * V + B * V) * TK**0.5) - P

    def MRK(self, P, TK):  # Redlich-Kwong routine to estimate endmember H2O and CO2 fugacities
        R = 83.14321
        B_1 = 14.6
        B_2 = 29.7

        # X_1 = 1
        for X_1 in [0, 1]:
            B = X_1 * B_1 + (1 - X_1) * B_2
            A = (X_1**2 * self.FNA(TK) + 2 * X_1 * (1 - X_1) * self.FNC(TK) +
                 (1 - X_1)**2 * self.FNB(TK))
            Temp2 = B + 5
            Q = 1
            Temp1 = 0
            while abs(Temp2 - Temp1) >= 0.00001:
                Temp1 = Temp2
                F_1 = (self.FNF(Temp1 + 0.01, TK, A, B, P) - self.FNF(Temp1, TK, A, B, P)) / 0.01
                Temp2 = Temp1 - Q * self.FNF(Temp1, TK, A, B, P) / F_1
                F_2 = (self.FNF(Temp2 + 0.01, TK, A, B, P) - self.FNF(Temp2, TK, A, B, P)) / 0.01
                if F_2 * F_1 <= 0:
                    Q = Q / 2.
                if abs(Temp2 - Temp1) > 0.00001:
                    F_1 = F_2
            V = Temp2
            G_1 = (np.log(V / (V - B)) + B_1 / (V - B) - 2 * (X_1 * self.FNA(TK) +
                   (1 - X_1) * self.FNC(TK)) * np.log((V + B) / V) / (R * TK**1.5 * B))
            G_1 = (G_1 + (np.log((V + B) / V) - B / (V + B)) * A * B_1 / (R * TK**1.5 * B**2) -
                   np.log(P * V / (R * TK)))
            G_1 = np.exp(G_1)
            G_2 = (np.log(V / (V - B)) + B_2 / (V - B) - 2 * (X_1 * self.FNC(TK) + (1 - X_1)
                   * self.FNB(TK)) * np.log((V + B) / V) / (R * TK**1.5 * B))
            G_2 = (G_2 + (np.log((V + B) / V) - B / (V + B)) * A * B_2 / (R * TK**1.5 * B**2) -
                   np.log(P * V / (R * TK)))
            G_2 = np.exp(G_2)
            if X_1 == 1:
                fH2Oo = G_1 * P  # The fugacity of H2O
                # return fH2Oo
        return fH2Oo


class fugacity_HB_co2(FugacityModel):
    """
    Implementation of the Holloway and Blank (1994) Modified Redlich Kwong EoS for CO2.
    """
    def __init__(self):
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', [1, 1e5], calibration_checks.crf_Between, 'bar', 'Redlich Kwong EOS',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'temperature', 500.0, calibration_checks.crf_GreaterThan, 'oC',
                'Redlich Kwong EOS',
                fail_msg=calibration_checks.crmsg_GreaterThan_fail,
                pass_msg=calibration_checks.crmsg_GreaterThan_pass,
                description_msg=calibration_checks.crmsg_GreaterThan_description)])
        self.HBmodel = fugacity_HollowayBlank()

    def fugacity(self, pressure, temperature, X_fluid=1.0, **kwargs):
        pure_f = self.HBmodel.fugacity(pressure=pressure, temperature=temperature, species='CO2')
        return pure_f * X_fluid


class fugacity_HB_h2o(FugacityModel):
    """
    Implementation of the Holloway and Blank (1994) Modified Redlich Kwong EoS for H2O.
    """
    def __init__(self):
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', [1, 1e5], calibration_checks.crf_Between, 'bar', 'Redlich Kwong EOS',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'temperature', 500.0, calibration_checks.crf_GreaterThan, 'oC',
                'Redlich Kwong EOS',
                fail_msg=calibration_checks.crmsg_GreaterThan_fail,
                pass_msg=calibration_checks.crmsg_GreaterThan_pass,
                description_msg=calibration_checks.crmsg_GreaterThan_description)])
        self.HBmodel = fugacity_HollowayBlank()

    def fugacity(self, pressure, temperature, X_fluid=1.0, **kwargs):
        pure_f = self.HBmodel.fugacity(pressure=pressure, temperature=temperature, species='H2O')
        return pure_f * X_fluid


class fugacity_HollowayBlank(FugacityModel):
    """
    Implementation of the Modified Redlich Kwong presented in Holloway and Blank (1994) Reviews
    in Mineralogy and Geochemistry vol. 30. Originally written in Quickbasic. CO2 calculations
    translated to Matlab by Chelsea Allison and translated to python by K. Iacovino for VESIcal.
    H2O calculations translated to VisualBasic by Gordon M. Moore and translated to python by
    K. Iacovino for VESIcal.

    """

    def __init__(self):
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', [1, 1e5], calibration_checks.crf_Between, 'bar',
                'MRK EOS (Holloway and Blank, 1994)',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'temperature', 500, calibration_checks.crf_GreaterThan, 'oC',
                'MRK EOS (Holloway and Blank, 1994)',
                fail_msg=calibration_checks.crmsg_GreaterThan_fail,
                pass_msg=calibration_checks.crmsg_GreaterThan_pass,
                description_msg=calibration_checks.crmsg_GreaterThan_description)])

    def REDKW(self, BP, A2B):
        """
        The RK routine. A routine to calculate compressibility factor and fugacity coefficient
        with the Redlich-Kwong equation following Edmister (1968). This solution for supercritical
        fluid.

        Parameters
        ----------
        BP: float
            B parameter sum from RKCALC

        A2B: float
            A parameter sum from RKCALC

        Returns
        -------
        float
            XLNFP (fugacity coefficient?)
        """
        if A2B < 1*10**(-10):
            A2B = 0.001

        # Define constants
        TH = 0.333333
        RR = -A2B*BP**2
        QQ = BP*(A2B-BP-1)
        XN = QQ*TH+RR-0.074074
        XM = QQ-TH
        XNN = XN*XN*0.25
        XMM = XM**3 / 27.0
        ARG = XNN+XMM

        if ARG > 0:
            X = np.sqrt(ARG)
            F = 1
            XN2 = -XN*0.5
            iXMM = XN2+X

            if iXMM < 0:
                F = -1
            XMM = F*((F*iXMM)**TH)
            F = 1
            iXNN = XN2 - X

            if iXNN < 0:
                F = -1
            XNN = F*((F*iXNN)**TH)
            Z = XMM+XNN+TH
            ZBP = Z-BP
            if ZBP < 0.000001:
                ZBP = 0.000001

            BPZ = 1+BP/Z
            FP = Z-1-np.log(ZBP)-A2B*np.log(BPZ)

            if FP < -37 or FP > 37:
                FP = 0.000001

        elif ARG < 0:
            COSPHI = np.sqrt(-XNN/XMM)
            if XN > 0:
                COSPHI = -COSPHI

            TANPHI = np.sqrt(1-COSPHI**2)/COSPHI
            PHI = np.arctan(TANPHI)*TH
            FAC = 2*np.sqrt(-XM*TH)

            # sort for largest root
            R1 = np.cos(PHI)
            R2 = np.cos(PHI+2.0944)
            R3 = np.cos(PHI+4.18879)
            RH = R2

            if R1 > R2:
                RH = R1
            if R3 > RH:
                RH = R3

            Z = RH*FAC+TH
            ZBP = Z-BP
            if ZBP < 0.000001:
                ZBP = 0.000001
            BPZ = 1+BP/Z
            FP = Z-1-np.log(ZBP)-A2B*np.log(BPZ)
            if FP < -37 or FP > 37:
                FP = 0.000001
        else:
            FP = 1
            Z = 1
        XLNFP = FP

        return XLNFP

    def Saxena(self, TK, pb):
        """
        High pressure corresponding states routines from Saxena and Fei (1987) GCA
        vol. 51, 783-791.

        Parameters
        ----------
        TK: float
            Temperature in K.

        pb: float
            Pressure in bars.

        Returns
        -------
        float
            XLNF, Natural log of the ratio F(P)/F(4000 bar)
        """

        # Define integration limit
        PO = 4000

        # Critical temperatures and pressures for CO2
        TR = TK/304.2
        PC = 73.9

        # Virial coeficients
        A = 2.0614-2.2351/TR**2 - 0.39411*np.log(TR)
        B = 0.055125/TR + 0.039344/TR**2
        C = -1.8935*10**(-6)/TR - 1.1092*10**(-5)/TR**2 - 2.1892*10**(-5)/TR**3
        D = 5.0527*10**(-11)/TR - 6.3033*10**(-21)/TR**3

        # integrate from PO (4000 bars) to P to calculate ln fugacity
        LNF = A*np.log(pb/PO)+(B/PC)*(pb-PO)+(C/(2*PC**2))*(pb**2-PO**2)
        LNF = LNF+(D/(3*PC**3))*(pb**3-PO**3)
        XLNF = LNF

        return XLNF

    def RKCALC(self, temperature, pressure, species):
        """
        Calculation of pure gas MRK properties following Holloway 1981, 1987

        Parameters
        ----------
        temperature: float
            Temperature in degrees K.

        pressure: float
            Pressure in atmospheres.

        Returns
        -------
        float
            Natural log of the fugacity of a pure gas.
        """
        # Define constants
        R = 82.05736
        pb = 1.013*pressure
        PBLN = np.log(pb)
        TCEL = temperature-273.15
        RXT = R*temperature
        RT = R*temperature**1.5 * 10**(-6)

        if species == 'CO2':
            # Calculate T-dependent MRK A parameter CO2
            ACO2M = 73.03 - 0.0714*TCEL + 2.157*10**(-5)*TCEL**2

            # Define MRK B parameter for CO2
            BSUM = 29.7

            ASUM = ACO2M / (BSUM*RT)

        elif species == 'H2O':
            # Calculate T-dependent MRK A parameter H2O
            AH2OM = 115.98 - np.double(0.0016295)*temperature - 1.4984*10**(-5)*temperature**2

            # Define MRK B parameter for H2O
            BSUM = 14.5

            ASUM = AH2OM / (BSUM*RT)

        BSUM = pressure*BSUM/RXT
        XLNFP = self.REDKW(BSUM, ASUM)

        # Convert to ln(fugacity)
        PUREG = XLNFP + PBLN
        return PUREG

    def fugacity(self, pressure, temperature, species, **kwargs):
        """
        Calculates fugacity.

        Parameters
        ----------
        temperature: float
            Temperature in degrees C.

        pressure: float
            Pressure in bars.

        species: str
            Choose which species to calculate. Options are 'H2O' and 'CO2'.

        Returns
        -------
        float
            Fugacity coefficient for passed species
        """

        # convert temp and press to atmospheres and Kelvin
        pressureAtmo = pressure/1.013
        temperatureK = temperature + 273.15
        PO = 4000/1.013

        # Use the MRK below 4,000 bars, Saxena above 4,000 bars
        if pressure > 4000 and species == 'CO2':
            iPUREG = self.RKCALC(temperatureK, PO, species)
            XLNF = self.Saxena(temperatureK, pressure)
            PUREG = iPUREG + XLNF
        else:
            PUREG = self.RKCALC(temperatureK, pressureAtmo, species)

        # Convert from ln(fugacity) to fugacity
        stdf = np.exp(PUREG)
        return stdf


class fugacity_RK_co2(FugacityModel):
    """
    Implementation of the Redlich Kwong EoS for CO2.
    Code derived from http://people.ds.cam.ac.uk/pjb10/thermo/pure.html - Patrick J. Barrie 30
    October 2003.
    """
    def __init__(self):
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', [1, 1e5], calibration_checks.crf_Between, 'bar', 'Redlich Kwong EOS',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'temperature', [500], calibration_checks.crf_GreaterThan, 'oC',
                'Redlich Kwong EOS',
                fail_msg=calibration_checks.crmsg_GreaterThan_fail,
                pass_msg=calibration_checks.crmsg_GreaterThan_pass,
                description_msg=calibration_checks.crmsg_GreaterThan_description)])

        self.RKmodel = fugacity_RedlichKwong()

    def fugacity(self, pressure, temperature, X_fluid, **kwargs):
        return self.RKmodel.fugacity(pressure, temperature, X_fluid, 'CO2')


class fugacity_RK_h2o(FugacityModel):
    """
    Implementation of the Redlich Kwong EoS for H2O.
    Code derived from http://people.ds.cam.ac.uk/pjb10/thermo/pure.html - Patrick J. Barrie 30
    October 2003.
    """
    def __init__(self):
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', [1, 1e5], calibration_checks.crf_Between, 'bar', 'Redlich Kwong EOS',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'temperature', 500, calibration_checks.crf_GreaterThan, 'oC', 'Redlich Kwong EOS',
                fail_msg=calibration_checks.crmsg_GreaterThan_fail,
                pass_msg=calibration_checks.crmsg_GreaterThan_pass,
                description_msg=calibration_checks.crmsg_GreaterThan_description)])
        self.RKmodel = fugacity_RedlichKwong()

    def fugacity(self, pressure, temperature, X_fluid, **kwargs):
        return self.RKmodel.fugacity(pressure, temperature, X_fluid, 'H2O')


class fugacity_RedlichKwong(FugacityModel):
    """
    Implementation of the Redlich Kwong EoS
    Code derived from http://people.ds.cam.ac.uk/pjb10/thermo/pure.html - Patrick J. Barrie 30
    October 2003.
    """
    def __init__(self):
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', [1, 1e5], calibration_checks.crf_Between, 'bar', 'Redlich Kwong EOS',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'temperature', 500, calibration_checks.crf_GreaterThan, 'oC', 'Redlich Kwong EOS',
                fail_msg=calibration_checks.crmsg_GreaterThan_fail,
                pass_msg=calibration_checks.crmsg_GreaterThan_pass,
                description_msg=calibration_checks.crmsg_GreaterThan_description)])

    def gamma(self, pressure, temperature, species):
        """
        Calculates fugacity coefficients.

        Parameters
        ----------
        temperature: fload
            Temperature in degrees C.

        pressure: float
            Pressure in bars.

        species: str
            Choose which species to calculate. Options are 'H2O' and 'CO2'.

        Returns
        -------
        float
            Fugacity coefficient for passed species.
        """

        temperatureK = temperature + 273.15
        R = 8.3145

        critical_params = {'CO2': {"cT":   304.15,
                                   "cP":   73.8659,
                                   "o":    0.225
                                   },
                           'H2O': {"cT":   647.25,
                                   "cP":   221.1925,
                                   "o":    0.334
                                   }
                           }

        # Calculate a and b parameters (depend only on critical parameters)...
        a = (0.42748 * R**2.0 * critical_params[species]["cT"]**(2.5) /
             (critical_params[species]["cP"] * 10.0**5))
        b = (0.08664 * R * critical_params[species]["cT"] /
             (critical_params[species]["cP"] * 10.0**5))

        # Calculate coefficients in the cubic equation of state...
        # coeffs: (C0, C1, C2, A, B)
        A = a * pressure * 10.0**5 / (np.sqrt(temperatureK) * (R * temperatureK)**2.0)
        B = b * pressure * 10.0**5 / (R * temperatureK)
        C2 = -1.0
        C1 = A - B - B * B
        C0 = -A * B

        # Solve the cubic equation for Z0 - Z2, D...
        Q1 = C2 * C1 / 6.0 - C0 / 2.0 - C2**3.0 / 27.0
        P1 = C2**2.0 / 9.0 - C1 / 3.0
        D = Q1**2.0 - P1**3.0

        if D >= 0:
            kOneThird = 1.0 / 3.0

            absQ1PSqrtD = np.fabs(Q1 + np.sqrt(D))
            temp1 = absQ1PSqrtD**kOneThird
            temp1 *= (Q1 + np.sqrt(D)) / absQ1PSqrtD

            absQ1MSqrtD = np.fabs(Q1 - np.sqrt(D))
            temp2 = absQ1MSqrtD**kOneThird
            temp2 *= (Q1 - np.sqrt(D)) / absQ1MSqrtD

            Z0 = temp1 + temp2 - C2 / 3.0
        else:
            temp1 = Q1**2.0 / (P1**3.0)
            temp2 = np.sqrt(1.0 - temp1) / np.sqrt(temp1)
            temp2 *= Q1 / np.fabs(Q1)

            gamma = np.arctan(temp2)

            if gamma < 0:
                gamma = gamma + np.pi

            Z0 = 2.0 * np.sqrt(P1) * np.cos(gamma/3.0) - C2 / 3.0
            Z1 = 2.0 * np.sqrt(P1) * np.cos((gamma + 2.0 * np.pi) / 3.0) - C2/3.0
            Z2 = 2.0 * np.sqrt(P1) * np.cos((gamma + 4.0 * np.pi) / 3.0) - C2/3.0

            if Z0 < Z1:
                temp0 = Z0
                Z0 = Z1
                Z1 = temp0

            if Z1 < Z2:
                temp0 = Z1
                Z1 = Z2
                Z2 = temp0

            if Z0 < Z1:
                temp0 = Z0
                Z0 = Z1
                Z1 = temp0

        # Calculate Departure Functions
        gamma = np.exp(Z0 - 1.0 - np.log(Z0-B) - A * np.log(1.0+B/Z0)/B)

        return gamma

    def fugacity(self, pressure, temperature, X_fluid=1.0, species='H2O', **kwargs):
        """
        Calculates the fugacity of H2O in a mixed H2O-CO2 fluid using the universal relationships:
        P_i = f_i/gamma_i = (fpure_i * Xfluid_i) / gamma_i
        See Iacovino (2015) EPSL for further explanation.
        """

        gammaH2O = self.gamma(pressure, temperature, 'H2O')
        gammaCO2 = self.gamma(pressure, temperature, 'CO2')

        fugacityH2Opure = pressure * gammaH2O
        fugacityCO2pure = pressure * gammaCO2

        if species == 'H2O':
            return fugacityH2Opure * X_fluid
        elif species == 'CO2':
            return fugacityCO2pure * X_fluid
        else:
            raise core.InputError("Species must be H2O or CO2.")



from VESIcal import core
from VESIcal import calibrations
from VESIcal.tasplot import add_LeMaitre_fields

import pandas as pd
import numpy as np
import warnings as w
import matplotlib as mpl
import matplotlib.pyplot as plt


# ---------- DEFINE CUSTOM PLOTTING FORMATTING ------------ #
style = "seaborn-colorblind"
plt.style.use(style)
plt.rcParams["mathtext.default"] = "regular"
plt.rcParams["mathtext.fontset"] = "dejavusans"
mpl.rcParams['patch.linewidth'] = 1
mpl.rcParams['axes.linewidth'] = 1
plt.rcParams['axes.titlesize'] = 20
plt.rcParams['axes.labelsize'] = 18
plt.rcParams['xtick.labelsize'] = 14
plt.rcParams['ytick.labelsize'] = 14
plt.rcParams['legend.fontsize'] = 14
mpl.rcParams['lines.markersize'] = 10

# Define color cycler based on plot style set here
# get style formatting set by plt.style.use():
the_rc = plt.style.library[style]
# list of colors by hex code:
color_list = the_rc['axes.prop_cycle'].by_key()['color'] * 10
color_cyler = the_rc['axes.prop_cycle']  # get the cycler


# ----------- MAGMASAT PLOTTING FUNCTIONS ----------- #
def smooth_isobars_and_isopleths(isobars=None, isopleths=None):
    """
    Takes in a dataframe with calculated isobar and isopleth information
    (e.g., output from calculate_isobars_and_isopleths) and smooths the data
    for plotting.

    Parameters
    ----------
    isobars: pandas DataFrame
        OPTIONAL. DataFrame object containing isobar information as calculated
        by calculate_isobars_and_isopleths.

    isopleths: pandas DataFrame
        OPTIONAL. DataFrame object containing isopleth information as
        calculated by calculate_isobars_and_isopleths.

    Returns
    -------
    pandas DataFrame
        DataFrame with x and y values for all isobars and all isopleths.
        Useful if a user wishes to do custom plotting with isobar and isopleth
        data rather than using the built-in `plot_isobars_and_isopleths()`
        function.
    """
    np.seterr(divide='ignore', invalid='ignore')  # turn off numpy warning
    w.filterwarnings("ignore", message="Polyfit may be poorly conditioned")

    if isobars is not None:
        P_vals = isobars.Pressure.unique()
        isobars_lists = isobars.values.tolist()
        # add zero values to volatiles list
        isobars_lists.append([0.0, 0.0, 0.0, 0.0])

        isobars_pressure = []
        isobars_H2O_liq = []
        isobars_CO2_liq = []
        # do some data smoothing
        for pressure in P_vals:
            Pxs = [item[1] for item in isobars_lists if item[0] == pressure]
            Pys = [item[2] for item in isobars_lists if item[0] == pressure]

            try:
                # calcualte polynomial
                Pz = np.polyfit(Pxs, Pys, 3)
                Pf = np.poly1d(Pz)

                # calculate new x's and y's
                Px_new = np.linspace(Pxs[0], Pxs[-1], 50)
                Py_new = Pf(Px_new)

                # Save x's and y's
                Px_new_list = list(Px_new)
                isobars_H2O_liq += Px_new_list

                Py_new_list = list(Py_new)
                isobars_CO2_liq += Py_new_list

                pressure_vals_for_list = [pressure]*len(Px_new)
                isobars_pressure += pressure_vals_for_list

            except Exception:
                Px_list = list(Pxs)
                isobars_H2O_liq += Px_list

                Py_list = list(Pys)
                isobars_CO2_liq += Py_list

                pressure_vals_for_list = [pressure]*len(Pxs)
                isobars_pressure += pressure_vals_for_list

        isobar_df = pd.DataFrame({"Pressure": isobars_pressure,
                                  "H2O_liq": isobars_H2O_liq,
                                  "CO2_liq": isobars_CO2_liq})

    if isopleths is not None:
        XH2O_vals = isopleths.XH2O_fl.unique()
        isopleths_lists = isopleths.values.tolist()

        isopleths_XH2O_fl = []
        isopleths_H2O_liq = []
        isopleths_CO2_liq = []
        for Xfl in XH2O_vals:
            Xxs = [item[1] for item in isopleths_lists if item[0] == Xfl]
            Xys = [item[2] for item in isopleths_lists if item[0] == Xfl]

            try:
                # calculate polynomial
                Xz = np.polyfit(Xxs, Xys, 2)
                Xf = np.poly1d(Xz)

                # calculate new x's and y's
                Xx_new = np.linspace(Xxs[0], Xxs[-1], 50)
                Xy_new = Xf(Xx_new)

                # Save x's and y's
                Xx_new_list = list(Xx_new)
                isopleths_H2O_liq += Xx_new_list

                Xy_new_list = list(Xy_new)
                isopleths_CO2_liq += Xy_new_list

                XH2Ofl_vals_for_list = [Xfl]*len(Xx_new)
                isopleths_XH2O_fl += XH2Ofl_vals_for_list

            except Exception:
                Xx_list = list(Xxs)
                isopleths_H2O_liq += Xx_list

                Xy_list = list(Xys)
                isopleths_CO2_liq += Xy_list

                XH2Ofl_vals_for_list = [Xfl]*len(Xxs)
                isopleths_XH2O_fl += XH2Ofl_vals_for_list

        isopleth_df = pd.DataFrame({"XH2O_fl": isopleths_XH2O_fl,
                                    "H2O_liq": isopleths_H2O_liq,
                                    "CO2_liq": isopleths_CO2_liq})

    np.seterr(divide='warn', invalid='warn')  # turn numpy warning back on
    w.filterwarnings("always", message="Polyfit may be poorly conditioned")

    if isobars is not None:
        if isopleths is not None:
            return isobar_df, isopleth_df
        else:
            return isobar_df
    else:
        if isopleths is not None:
            return isopleth_df


def plot(isobars=None, isopleths=None, degassing_paths=None, custom_H2O=None,
         custom_CO2=None, isobar_labels=None, isopleth_labels=None,
         degassing_path_labels=None, custom_labels=None,
         custom_colors="VESIcal", custom_symbols=None, markersize=10,
         figsize=(12, 8), save_fig=False, extend_isobars_to_zero=True,
         smooth_isobars=False, smooth_isopleths=False, **kwargs):
    """
    Custom automatic plotting of model calculations in VESIcal.
    Isobars, isopleths, and degassing paths can be plotted. Labels can be
    specified for each. Any combination of isobars, isopleths, and degassing
    paths can be plotted.

    Parameters
    ----------
    isobars: pandas DataFrame or list
        OPTIONAL. DataFrame object containing isobar information as calculated
        by calculate_isobars_and_isopleths. Or a list of DataFrame objects.

    isopleths: pandas DataFrame or list
        OPTIONAL. DataFrame object containing isopleth information as
        calculated by calculate_isobars_and_isopleths. Or a list of DataFrame
        objects.

    degassing_paths: list
        OPTIONAL. List of DataFrames with degassing information as generated
        by calculate_degassing_path().

    custom_H2O: list
        OPTIONAL. List of groups of H2O values to plot as points. For example
        myfile.data['H2O'] is one group of H2O values. Must be passed with
        custom_CO2 and must be same length as custom_CO2.

    custom_CO2: list
        OPTIONAL. List of groups of CO2 values to plot as points.For example
        myfile.data['CO2'] is one group of CO2 values. Must be passed with
        custom_H2O and must be same length as custom_H2O.

    isobar_labels: list
        OPTIONAL. Labels for the plot legend. Default is None, in which case
        each plotted line will be given the generic legend name of
        "Isobars n", with n referring to the nth isobars passed. Isobar
        pressure is given in parentheses. The user can pass their own labels
        as a list of strings. If more than one set of isobars is passed, the
        labels should refer to each set of isobars, not each pressure.

    isopleth_labels: list
        OPTIONAL. Labels for the plot legend. Default is None, in which case
        each plotted isopleth will be given the generic legend name of
        "Isopleth n", with n referring to the nth isopleths passed. Isopleth
        XH2O values are given in parentheses. The user can pass their own
        labels as a list of strings. If more than one set of isopleths is
        passed, the labels should refer to each set of isopleths, not each
        XH2O value.

    degassing_path_labels: list
        OPTIONAL. Labels for the plot legend. Default is None, in which case
        each plotted line will be given the generic legend name of "Pathn",
        with n referring to the nth degassing path passed. The user can pass
        their own labels as a list of strings.

    custom_labels: list
        OPTIONAL. Labels for the plot legend. Default is None, in which case
        each group of custom points will be given the generic legend name of
        "Customn", with n referring to the nth degassing path passed. The user
        can pass their own labels as a list of strings.

    custom_colors: list
        OPTIONAL. Default value is "VESIcal", which uses VESIcal's color ramp.
        A list of color values readable by matplotlib can be passed here if
        custom symbol colors are desired. The length of this list must match
        that of custom_H2O and custom_CO2.

    custom_symbols: list
        OPTIONAL. Default value is None, in which case data are plotted as
        filled circles.. A list of symbol tyles readable by matplotlib can be
        passed here if custom symbol types are desired. The length of this
        list must match that of custom_H2O and custom_CO2.

    markersize: int
        OPTIONAL. Default value is 10. Same as markersize kwarg in matplotlib.
        Any numeric value passed here will set the marker size for
        (custom_H2O, custom_CO2) points.

    figsize: tuple
        OPTIONAL. Default value is (12,8). Sets the matplotlib.pyplot figsize
        value as (x_dimension, y_dimension)

    save_fig: False or str
        OPTIONAL. Default value is False, in which case the figure will not be
        saved. If a string is passed, the figure will be saved with the string
        as the filename. The string must include the file extension.

    extend_isobars_to_zero: bool
        OPTIONAL. If True (default), isobars will be extended to zero, even if
        there is a finite solubility at zero partial pressure.

    smooth_isobars: bool
        OPTIONAL. Default is False. If set to True, isobar data will be fit to
        a polynomial and plotted. If False, the raw input data will be plotted.

    smooth_isopleths: bool
        OPTIONAL. Default is False. If set to True, isopleth data will be fit
        to a polynomial and plotted. If False, the raw input data will be
        plotted.

    Returns
    -------
    fig, axes Matplotlib objects
        fig and axes matploblib objects defining a plot with x-axis as H2O wt%
        in the melt and y-axis as CO2 wt%in the melt. Isobars, or lines of
        constant pressure at which the sample magma composition is saturated,
        and isopleths, or lines of constant fluid composition at which the
        sample magma composition is saturated, are plotted if passed.
        Degassing paths, or the concentration of dissolved H2O and CO2 in a
        melt equilibrated along a path of decreasing pressure, is plotted if
        passed.
    """
    # Turn off warnings:
    np.seterr(divide='ignore', invalid='ignore')  # turn off numpy warning
    w.filterwarnings("ignore", message="Polyfit may be poorly conditioned")

    def check_inputs(custom_H2O, custom_CO2):
        if custom_H2O is not None:
            if custom_CO2 is None:
                raise core.InputError("If x data is passed, y data must also "
                                      "be passed.")
            else:
                if len(custom_H2O) == len(custom_CO2):
                    pass
                else:
                    raise core.InputError("x and y data must be same length")
        if custom_CO2 is not None:
            if custom_H2O is None:
                raise core.InputError("If y data is passed, x data must also "
                                      "be passed.")

    def check_colors(custom_colors):
        if custom_colors == "VESIcal":
            use_colors = color_list
        elif isinstance(custom_colors, list):
            use_colors = custom_colors
        else:
            raise core.InputError("Argument custom_colors must be type list. "
                                  "Just passing one item? Try putting square "
                                  "brackets, [], around it.")
        return use_colors

    def calc_extend_isobars_to_zero(Pxs, Pys):
        """
        Calculates new end-points for plotting isobars when
        extend_isobars_to_zero option is set to True.

        Parameters
        ----------
        Pxs, Pys: list
            List of x and y values corresponding to isobars.
        """
        if Pxs[0]*Pys[0] != 0.0:
            if Pxs[0] > Pys[0]:
                # create new array of length n+1 if n is the length of the
                # original array:
                Px_new = np.zeros(np.shape(Pxs)[0]+1)

                # set the first x value in the new array equal to 0:
                Px_new[0] = 0

                # fill the rest of the new array with the original array
                # values:
                Px_new[1:] = Pxs

                # overwrite the original array with the new one:
                Pxs = Px_new

                Py_new = np.zeros(np.shape(Pys)[0]+1)
                Py_new[0] = Pys[0]
                Py_new[1:] = Pys
                Pys = Py_new
            else:
                Px_new = np.zeros(np.shape(Pxs)[0]+1)
                Px_new[0] = Pxs[0]
                Px_new[1:] = Pxs
                Pxs = Px_new

                Py_new = np.zeros(np.shape(Pys)[0]+1)
                Py_new[0] = 0
                Py_new[1:] = Pys
                Pys = Py_new

        if Pxs[-1]*Pys[-1] != 0.0:
            if Pxs[-1] < Pys[-1]:
                Px_new = np.zeros(np.shape(Pxs)[0]+1)
                Px_new[-1] = 0
                Px_new[:-1] = Pxs
                Pxs = Px_new

                Py_new = np.zeros(np.shape(Pys)[0]+1)
                Py_new[-1] = Pys[-1]
                Py_new[:-1] = Pys
                Pys = Py_new
            else:
                Px_new = np.zeros(np.shape(Pxs)[0]+1)
                Px_new[-1] = Pxs[-1]
                Px_new[:-1] = Pxs
                Pxs = Px_new

                Py_new = np.zeros(np.shape(Pys)[0]+1)
                Py_new[-1] = 0
                Py_new[:-1] = Pys
                Pys = Py_new

        return Pxs, Pys

    # -------- HANDLE USER INPUT ERRORS, SET COLORS, SMOOTH LINES -------- ##
    check_inputs(custom_H2O=custom_H2O, custom_CO2=custom_CO2)
    use_colors = check_colors(custom_colors=custom_colors)

    if smooth_isobars:
        isobars = smooth_isobars_and_isopleths(isobars=isobars)
    if smooth_isopleths:
        isopleths = smooth_isobars_and_isopleths(isopleths=isopleths)

    # -------- CREATE FIGURE -------- ##
    fig, ax = plt.subplots(figsize=figsize)
    if 'custom_x' in kwargs:
        ax.set(xlabel=kwargs['xlabel'], ylabel=kwargs['ylabel'])
    else:
        ax.set(xlabel='H$_2$O wt%', ylabel='CO$_2$ wt%')

    labels = []

    # -------- PLOT ISOBARS -------- ##
    if isobars is not None:
        if isinstance(isobars, pd.DataFrame):
            isobars = [isobars]

        for i in range(len(isobars)):
            P_vals = isobars[i].Pressure.unique()
            isobars_lists = isobars[i].values.tolist()

            # add zero values to volatiles list
            isobars_lists.append([0.0, 0.0, 0.0, 0.0])

            P_iter = 0
            for pressure in P_vals:
                P_iter += 1
                Pxs = [item[1] for item in isobars_lists
                       if item[0] == pressure]
                Pys = [item[2] for item in isobars_lists
                       if item[0] == pressure]

                if extend_isobars_to_zero:
                    try:
                        Pxs, Pys = calc_extend_isobars_to_zero(Pxs, Pys)
                    except Exception:
                        pass
                else:
                    print(extend_isobars_to_zero)

                if len(isobars) > 1:
                    if P_iter == 1:
                        P_list = [int(i) for i in P_vals]
                        if isinstance(isobar_labels, list):
                            labels.append(str(isobar_labels[i]) + ' (' +
                                          ', '.join(map(str, P_list)) +
                                          " bars)")
                        else:
                            labels.append('Isobars ' + str(i+1) + ' (' +
                                          ', '.join(map(str, P_list)) +
                                          " bars)")
                    else:
                        labels.append('_nolegend_')

                if len(isobars) > 1:
                    ax.plot(Pxs, Pys, color=color_list[i])
                else:
                    ax.plot(Pxs, Pys)

            if len(isobars) == 1:
                labels += [str(P_val) + " bars" for P_val in P_vals]

    # -------- PLOT ISOPLETHS -------- ##
    if isopleths is not None:
        if isinstance(isopleths, pd.DataFrame):
            isopleths = [isopleths]

        for i in range(len(isopleths)):
            XH2O_vals = isopleths[i].XH2O_fl.unique()
            isopleths_lists = isopleths[i].values.tolist()

            H_iter = 0
            for Xfl in XH2O_vals:
                H_iter += 1
                Xxs = [item[1] for item in isopleths_lists if item[0] == Xfl]
                Xys = [item[2] for item in isopleths_lists if item[0] == Xfl]

                if len(isopleths) > 1:
                    if H_iter == 1:
                        H_list = [i for i in XH2O_vals]
                        if isinstance(isopleth_labels, list):
                            labels.append(str(isopleth_labels[i]) + ' (' +
                                          ', '.join(map(str, H_list)) +
                                          " XH2Ofluid)")
                        else:
                            labels.append('Isopleths ' + str(i+1) + ' (' +
                                          ', '.join(map(str, H_list)) +
                                          " XH2Ofluid)")
                    else:
                        labels.append('_nolegend_')
                    ax.plot(Xxs, Xys, ls='dashed', color=color_list[i])

                if len(isopleths) == 1:
                    H_list = [i for i in XH2O_vals]
                    if H_iter == 1:
                        labels.append('Isopleths (' +
                                      ', '.join(map(str, H_list)) +
                                      " XH2Ofluid)")
                    else:
                        labels.append('_nolegend_')
                    ax.plot(Xxs, Xys, ls='dashed', color='k')

    # -------- PLOT DEGASSING PATHS -------- ##
    if degassing_paths is not None:
        if isinstance(degassing_paths, pd.DataFrame):
            degassing_paths = [degassing_paths]

        degassing_colors = color_list.copy()
        iterno = 0
        plot_type = None
        for i in range(len(degassing_paths)):
            # handle whether labels are passed
            if degassing_path_labels is None:
                iterno += 1
                labels.append('Path%s' % iterno)
            else:
                labels.append(degassing_path_labels[iterno])
                iterno += 1

            # handle if only one volatile present
            if (degassing_paths[i]["H2O_liq"].max() > 0 and
                    degassing_paths[i]["CO2_liq"].max() > 0):
                ax.plot(degassing_paths[i]["H2O_liq"],
                        degassing_paths[i]["CO2_liq"],
                        ls='dotted', color=degassing_colors[i])
                ax.plot(degassing_paths[i]["H2O_liq"].max(),
                        degassing_paths[i]["CO2_liq"].max(),
                        'o', color=degassing_colors[i])
                # warn user if trying to plot mixed types on same figure
                if plot_type not in [None, "mixed"]:
                    w.warn("Warning: you are trying to plot two different plot types on the same"
                           " figure. Curves may not match axis labels.", stacklevel=2)
                plot_type = "mixed"
            elif (degassing_paths[i]["H2O_liq"].max() > 0 and
                  degassing_paths[i]["CO2_liq"].max() <= 0):
                ax.plot(degassing_paths[i]["Pressure_bars"],
                        degassing_paths[i]["H2O_liq"],
                        ls='dotted', color=degassing_colors[i])
                ax.plot(degassing_paths[i]["Pressure_bars"].max(),
                        degassing_paths[i]["H2O_liq"].max(),
                        'o', color=degassing_colors[i])
                ax.set(xlabel='Pressure (bars)', ylabel='H$_2$O wt%')
                # warn user if trying to plot mixed types on same figure
                if plot_type not in [None, "H2O"]:
                    w.warn("Warning: you are trying to plot two different plot types on the same"
                           " figure. Curves may not match axis labels.", stacklevel=2)
                plot_type = "H2O"
            elif (degassing_paths[i]["H2O_liq"].max() <= 0 and
                  degassing_paths[i]["CO2_liq"].max() > 0):
                ax.plot(degassing_paths[i]["Pressure_bars"],
                        degassing_paths[i]["CO2_liq"],
                        ls='dotted', color=degassing_colors[i])
                ax.plot(degassing_paths[i]["Pressure_bars"].max(),
                        degassing_paths[i]["CO2_liq"].max(),
                        'o', color=degassing_colors[i])
                ax.set(xlabel='Pressure (bars)', ylabel='CO$_2$ wt%')
                # warn user if trying to plot mixed types on same figure
                if plot_type not in [None, "CO2"]:
                    w.warn("Warning: you are trying to plot two different plot types on the same"
                           " figure. Curves may not match axis labels.", stacklevel=2)
                plot_type = "CO2"

            labels.append('_nolegend_')

    # -------- PLOT CUSTOM H2O-CO2 -------- ##
    if custom_H2O is not None and custom_CO2 is not None:
        if isinstance(custom_H2O, pd.DataFrame):
            custom_H2O = [custom_H2O]
        if isinstance(custom_CO2, pd.DataFrame):
            custom_CO2 = [custom_CO2]

        if custom_symbols is None:
            use_marker = ['o'] * len(custom_H2O)
        else:
            use_marker = custom_symbols

        iterno = 0
        for i in range(len(custom_H2O)):
            if custom_labels is None:
                iterno += 1
                labels.append('Custom%s' % iterno)
                ax.plot(custom_H2O[i], custom_CO2[i], use_marker[i],
                        color=use_colors[i], markersize=markersize)
            else:
                labels.append(custom_labels[iterno])
                ax.plot(custom_H2O[i], custom_CO2[i], use_marker[i],
                        color=use_colors[i], markersize=markersize)
                iterno += 1

    # -------- PLOT CUSTOM X-Y -------- ##
    if 'custom_x' in kwargs:
        custom_x = kwargs['custom_x']
        custom_y = kwargs['custom_y']

        if isinstance(custom_x, pd.core.series.Series):
            custom_x = [list(custom_x.values)]
        if isinstance(custom_y, pd.core.series.Series):
            custom_y = [list(custom_y.values)]

        if custom_symbols is None:
            use_marker = ['o'] * len(custom_x)
        else:
            use_marker = custom_symbols

        iterno = 0
        for i in range(len(custom_x)):
            if custom_labels is None:
                iterno += 1
                labels.append('Custom%s' % iterno)
                ax.plot(custom_x[i], custom_y[i], use_marker[i],
                        color=use_colors[i], markersize=markersize)
            else:
                labels.append(custom_labels[iterno])
                ax.plot(custom_x[i], custom_y[i], use_marker[i],
                        color=use_colors[i], markersize=markersize)
                iterno += 1

    # -------- PLOT LEGEND -------- ##
    ax.legend(labels, bbox_to_anchor=(1.01, 1), loc='upper left')

    if 'custom_x' not in kwargs:
        ax.set_xlim(left=0)
        ax.set_ylim(bottom=0)

    np.seterr(divide='warn', invalid='warn')  # turn numpy warning back on
    w.filterwarnings("always", message="Polyfit may be poorly conditioned")

    # -------- SAVE FIGURE IF DESIRED -------- ##
    if save_fig is not False:
        fig.savefig(save_fig)

    return fig, ax


def scatterplot(custom_x, custom_y, xlabel=None, ylabel=None, **kwargs):
    """
    Custom x-y plotting using VESIcal's built-in plot() function, built
    Matplotlib's plot and scatter functions.

    Parameters
    ----------
    custom_x: list
        List of groups of x-values to plot as points or lines

    custom_y: list
        List of groups of y-values to plot as points or lines

    xlabel: str
        OPTIONAL. What to display along the x-axis.

    ylabel: str
        OPTIONAL. What to display along the y-axis.

    kwargs:
        Can take in any key word agruments that can be passed to `plot()`.

    Returns
    -------
    fig, ax matplotlib objects
        X-y plot with custom x and y axis values and labels.
    """

    if isinstance(custom_x, list) and isinstance(custom_y, list):
        if len(custom_x) != len(custom_y):
            raise core.InputError("X and y lists must be same length")

    if xlabel is not None:
        if isinstance(xlabel, str):
            pass
        else:
            raise core.InputError("xlabel must be string")

    if ylabel is not None:
        if isinstance(ylabel, str):
            pass
        else:
            raise core.InputError("ylabel must be string")

    return plot(custom_x=custom_x, custom_y=custom_y, xlabel=xlabel,
                ylabel=ylabel, **kwargs)


# ------- Define custom plotting tools for checking calibrations ------- #

def calib_plot(user_data=None, model='all', plot_type='TAS', zoom=None,
               figsize=(17, 8), legend=True, save_fig=False, **kwargs):
    """
    Plots user data and calibration set of any or all models on any x-y plot
    or a total alkalis vs silica (TAS) diagram. TAS diagram boundaries
    provided by tasplot python module, copyright John A Stevenson.

    Parameters
    ----------
    user_data: BatchFile object, Sample object, pandas DataFrame, pandas Series,
        or dict.
        OPTIONAL. Default value is None, in which case only the model
        calibration set is plotted. User provided sample data describing the
        oxide composition of one or more samples. Multiple samples can be
        passed as an BatchFile object or pandas DataFrame. A single sample can
        be passed as a pandas Series.

    model: str or list
        OPTIONAL. Default value is 'all', in which case all model calibration
        datasets will be plotted. 'Mixed' can be used to plot all mixed fluid
        models. String of the name of the model calibration dataset to plot
        (e.g., 'Shishkina'). Multiple models can be plotted by passing them as
        strings within a list (e.g., ['Shishkina', 'Dixon']).

    plot_type: str
        OPTIONAL. Default value is 'TAS', which returns a total alkali vs
        silica (TAS) diagram. Any two oxides can be plotted as an x-y plot by
        setting plot_type='xy' and specifying x- and y-axis oxides, e.g.,
        x='SiO2', y='Al2O3'.

    zoom: str or list
        OPTIONAL. Default value is None in which case axes will be set to the
        default of 35<x<100 wt% and 0<y<25 wt% for TAS type plots and the best
        values to show the data for xy type plots. Can pass "user_data" to
        plot the figure where the x and y axes are scaled down to zoom in and
        only show the region surrounding the user_data. A list of tuples may
        be passed to manually specify x and y limits. Pass in data as
        [(x_min, x_max), (y_min, y_max)]. For example, the default limits here
        would be passed in as [(35,100), (0,25)].

    figsize: tuple
        OPTIONAL. Default value is (17,8). Sets the matplotlib.pyplot figsize
        value as (x_dimension, y_dimension).

    legend: bool
        OPTIONAL. Default value is True. Can be set to False in which case the
        legend will not be displayed.

    save_fig: False or str
        OPTIONAL. Default value is False, in which case the figure will not be
        saved. If a string is passed, the figure will be saved with the string
        as the filename. The string must include the file extension.

    Returns
    -------
    matplotlib object
    """

    # Get x and y axis limits, if user passed them
    if zoom is None:
        user_xmin = 35
        user_xmax = 100
        user_ymin = 0
        user_ymax = 25
    elif zoom == 'user_data':
        if isinstance(user_data, pd.DataFrame):
            print("'user_data' type zoom for more than one sample is not ",
                  "implemented yet.")
            user_xmin = 35
            user_xmax = 100
            user_ymin = 0
            user_ymax = 25
        elif (isinstance(user_data, pd.core.series.Series) or
              isinstance(user_data, dict)):
            user_xmin = user_data['SiO2'] - 5
            user_xmax = user_data['SiO2'] + 5
            user_ymin = user_data['Na2O'] + user_data['K2O'] - 2
            if user_ymin < 0:
                user_ymin = 0
            user_ymax = user_data['Na2O'] + user_data['K2O'] + 2
    elif isinstance(zoom, list):
        user_xmin, user_xmax = zoom[0]
        user_ymin, user_ymax = zoom[1]
    else:
        raise core.InputError('Trying to pass zoom coords? Pass as ' +
                              '[(x, x), (y, y)]')

    # Create the figure
    fig, ax1 = plt.subplots(figsize=figsize)
    font = {'family': 'sans-serif',
            'color':  'black',
            'weight': 'normal',
            'size':   20,
            }

    # TAS figure
    if plot_type == 'TAS':
        # adjust x limits here if you want to focus on a specific part of
        # compostional space:
        ax1.set_xlim([user_xmin, user_xmax])

        # adjust y limits here
        ax1.set_ylim([user_ymin, user_ymax])
        plt.xlabel('SiO$_2$, wt%', fontdict=font, labelpad=15)
        plt.ylabel('Na$_2$O+K$_2$O, wt%', fontdict=font, labelpad=15)

        # add LeMaitre fields
        if zoom is None:
            add_LeMaitre_fields(ax1)

    elif plot_type == 'xy':
        if 'x' in kwargs and 'y' in kwargs:
            x = kwargs['x']
            y = kwargs['y']
            if zoom is not None:
                ax1.set_xlim([user_xmin, user_xmax])
                ax1.set_ylim([user_ymin, user_ymax])
            plt.xlabel(str(x)+", wt%", fontdict=font, labelpad=15)
            plt.ylabel(str(y)+", wt%", fontdict=font, labelpad=15)
        else:
            raise core.InputError("If plot_type is 'xy', then x and y "
                                  "values must be passed as strings. For "
                                  "example, x='SiO2', y='Al2O3'.")

    # Plot Calibration Data
    if model == 'all':
        model = ['MagmaSat',
                 'Shishkina',
                 'Dixon',
                 'IaconoMarziano',
                 'Liu',
                 'AllisonCarbon',
                 'MooreWater']
    if model == 'mixed':
        model = ['MagmaSat',
                 'Shishkina',
                 'Dixon',
                 'IaconoMarziano',
                 'Liu']

    if isinstance(model, str):
        model = [model]

    if isinstance(model, list):
        # set legends to false
        h2o_legend = False
        co2_h2oco2_legend = False

        # check which legends to turn to True
        for modelname in model:
            model_type = calibrations.return_calibration_type(modelname)
            if model_type['H2O']:
                h2o_legend = True
            if model_type['CO2'] or model_type['Mixed']:
                co2_h2oco2_legend = True

        if h2o_legend:
            plt.scatter([], [], marker='', label=r"$\bf{Pure \ H_2O:}$")

            for modelname in model:
                calibdata = calibrations.return_calibration(modelname)
                model_type = calibrations.return_calibration_type(modelname)
                if isinstance(calibdata, str):
                    w.warn(calibdata)
                else:
                    if model_type['H2O']:
                        if plot_type == 'TAS':
                            try:
                                plt.scatter(calibdata['H2O']['SiO2'],
                                            (calibdata['H2O']['Na2O'] +
                                                calibdata['H2O']['K2O']),
                                            marker='s', edgecolors='k',
                                            facecolors=calibdata['facecolor'],
                                            label=str(modelname))
                            except Exception:
                                plt.scatter(calibdata['H2O']['SiO2'],
                                            calibdata['H2O']['Na2O+K2O'],
                                            marker='s', edgecolors='k',
                                            facecolors=calibdata['facecolor'],
                                            label=str(modelname))
                        if plot_type == 'xy':
                            try:
                                plt.scatter(calibdata['H2O'][x],
                                            calibdata['H2O'][y],
                                            marker='s', edgecolors='k',
                                            facecolors=calibdata['facecolor'],
                                            label=str(modelname))
                            except Exception:
                                w.warn("The requested oxides were not found",
                                       "in the calibration dataset for " +
                                       str(modelname) + ".")

            if co2_h2oco2_legend:
                plt.scatter([], [], marker='', label=r"${\ }$")

        if co2_h2oco2_legend:
            plt.scatter([], [], marker='',
                        label=r"$\bf{\ CO_2 \ and \ H_2O\!-\!CO_2:}$")

        for modelname in model:
            calibdata = calibrations.return_calibration(modelname)
            model_type = calibrations.return_calibration_type(modelname)
            if isinstance(calibdata, str):
                w.warn(calibdata)
            else:
                if model_type['CO2'] and model_type['Mixed']:
                    frames = [calibdata['CO2'], calibdata['Mixed']]
                    co2_and_mixed = pd.concat(frames)
                    if plot_type == 'TAS':
                        try:
                            plt.scatter(co2_and_mixed['SiO2'],
                                        (co2_and_mixed['Na2O'] +
                                        co2_and_mixed['K2O']),
                                        marker='d', edgecolors='k',
                                        facecolors=calibdata['facecolor'],
                                        label=str(modelname))
                        except Exception:
                            plt.scatter(co2_and_mixed['SiO2'],
                                        co2_and_mixed['Na2O+K2O'],
                                        marker='d', edgecolors='k',
                                        facecolors=calibdata['facecolor'],
                                        label=str(modelname))
                    if plot_type == 'xy':
                        try:
                            plt.scatter(co2_and_mixed[x], co2_and_mixed[y],
                                        marker='d', edgecolors='k',
                                        facecolors=calibdata['facecolor'],
                                        label=str(modelname))
                        except Exception:
                            w.warn("The requested oxides were not found in ",
                                   "the calibration dataset for " +
                                   str(modelname) + ".")
                elif model_type['CO2'] or model_type['Mixed']:
                    if model_type['CO2']:
                        thistype = 'CO2'
                    if model_type['Mixed']:
                        thistype = 'Mixed'
                    if plot_type == 'TAS':
                        try:
                            plt.scatter(calibdata[thistype]['SiO2'],
                                        (calibdata[thistype]['Na2O'] +
                                        calibdata[thistype]['K2O']),
                                        marker='d', edgecolors='k',
                                        facecolors=calibdata['facecolor'],
                                        label=str(modelname))
                        except Exception:
                            plt.scatter(calibdata[thistype]['SiO2'],
                                        calibdata[thistype]['Na2O+K2O'],
                                        marker='d', edgecolors='k',
                                        facecolors=calibdata['facecolor'],
                                        label=str(modelname))
                    if plot_type == 'xy':
                        try:
                            plt.scatter(calibdata[thistype][x],
                                        calibdata[thistype][y],
                                        marker='d', edgecolors='k',
                                        facecolors=calibdata['facecolor'],
                                        label=str(modelname))
                        except Exception:
                            w.warn("The requested oxides were not found in ",
                                   "the calibration dataset for "
                                   + str(modelname) + ".")
    else:
        raise core.InputError("model must be of type str or list")

    # Plot user data
    if user_data is None:
        pass
    else:
        if ((user_data.__class__.__module__, user_data.__class__.__name__) ==
           ('VESIcal', 'BatchFile')):
            user_data = user_data.get_data()
            # batchfile and VESIcal (__init__) are not imported to avoid
            # circular imports
            # use above notation to interrogate datatype
        if ((user_data.__class__.__module__, user_data.__class__.__name__) ==
           ('VESIcal', 'Sample')):
            user_data = user_data.get_composition()
            # batchfile and VESIcal (__init__) are not imported to avoid
            # circular imports
            # use above notation to interrogate datatype
        if plot_type == 'TAS':
            _sample = user_data.copy()
            try:
                _sample["TotalAlkalis"] = _sample["Na2O"] + _sample["K2O"]
            except Exception:
                core.InputError("Na2O and K2O data must be in user_data")
            plt.scatter(_sample['SiO2'], _sample['TotalAlkalis'],
                        s=150, edgecolors='w', facecolors='red', marker='P',
                        label='User Data')
        if plot_type == 'xy':
            _sample = user_data.copy()
            plt.scatter(_sample[x], _sample[y],
                        s=150, edgecolors='w', facecolors='red', marker='P',
                        label='User Data')

    if legend:
        plt.legend(bbox_to_anchor=(1.04, 1), loc="upper left")
    fig.tight_layout()
    if isinstance(save_fig, str):
        fig.savefig(save_fig)

    return fig, ax1


def show():
    """
    Local implementation of pyplot.show(). For displaying created plots.
    """
    plt.show()



from VESIcal import core

import numpy as np

giordano_oxide_mass = {'SiO2': 60.0843,
                       'TiO2': 79.8658,
                       'Al2O3': 101.961276,
                       'FeO': 71.8444,
                       'MnO': 70.937449,
                       'MgO': 40.3044,
                       'CaO': 56.0774,
                       'Na2O': 61.97894,
                       'K2O': 94.1960,
                       'P2O5': 141.9446,
                       'H2O': 18.01528,
                       'F2O': 18.9984*2}


def _normalize_Giordano(sample):
    """
    Converts an input, not necessarily normalized composition to a
    composition in mol% oxides, using the normalization routine in the Giordano
    et al. (2008) excel spreadsheet.

    Parameters
    ----------
    sample  Sample class
        Magma major element composition.

    Returns
    -------
    dict
        Normalized major element composition in mol% oxides.
    """

    _sample = sample.get_composition(units='wtpt_oxides',
                                     oxide_masses=giordano_oxide_mass,
                                     asSampleClass=True)

    # convert FeO and Fe2O3 to FeOT as FeO
    _sample_FeOT = (_sample.get_composition(units='wtpt_oxides',
                                            oxide_masses=giordano_oxide_mass,
                                            species='FeO') +
                    0.8998*_sample.get_composition(units='wtpt_oxides',
                                                   species='Fe2O3'))
    _sample.change_composition({'FeO': _sample_FeOT, 'Fe2O3': 0.0})

    # if passed, convert F to F2O
    _sample_F2O = _sample.get_composition(units='wtpt_oxides',
                                          oxide_masses=giordano_oxide_mass,
                                          species='F2O')
    if 'F' in _sample.get_composition().keys():
        _sample_F2O += (_sample.get_composition()['F'] *
                        giordano_oxide_mass['F2O']/core.oxideMass['F'])
    _sample.change_composition({'F2O': _sample_F2O})

    # ensure all necessary oxides are in sample
    oxides_giordano = ['SiO2', 'TiO2', 'Al2O3', 'FeO', 'MnO', 'MgO',
                       'CaO', 'Na2O', 'K2O', 'P2O5', 'H2O', 'F2O']
    sample_wtper = {}
    for oxide in oxides_giordano:
        if oxide in _sample.get_composition().keys():
            sample_wtper[oxide] = _sample.get_composition(units='wtpt_oxides',
                                                          oxide_masses=giordano_oxide_mass,
                                                          species=oxide)
        else:
            sample_wtper[oxide] = 0.0

    # special normalize wt%
    orig_sample_sum_anhy = (sum([v for v in sample_wtper.values()]) -
                            sample_wtper['H2O'])

    sample_wtper_norm = {}
    for oxide in sample_wtper.keys():
        if oxide == 'H2O':
            sample_wtper_norm[oxide] = sample_wtper['H2O']
        else:
            sample_wtper_norm[oxide] = ((100.0-sample_wtper['H2O']) *
                                        sample_wtper[oxide]/(orig_sample_sum_anhy))

    GFW = (100.0 /
           (sum([sample_wtper_norm[oxide]/core.oxideMass[oxide] for oxide
                in oxides_giordano])))

    sample_moloxides_giordano = {oxide: GFW*sample_wtper_norm[oxide] /
                                 giordano_oxide_mass[oxide] for oxide in
                                 oxides_giordano}

    return sample_moloxides_giordano


def calculate_liquid_viscosity(sample, temperature, **kwargs):
    """
    Calculates the viscosity of a liquid given the chemical composition and
    temperature.

    Parameters
    ----------
    sample  Sample class
        Magma major element composition.

    temperature     float or int
        Temperature in degrees C.

    Returns
    -------
    float
        Log viscosity of the liquid in Pa*s, rounded to 4 dp.
    """

    # normalize using Giordano et al. (2008) spreadsheet routine
    comp_molpercent = _normalize_Giordano(sample)

    # known A constant value (Table 1, Giordano et al., 2008)
    A_value = -4.55

    def _calculate_partial_B_value(bulk_comp, b_number):
        """
        Calculate values for B (computed values in Giordano
        spreadsheet).

        B = sum(i=1-7)[b_i*M_i] + sum(j=1-3)[b_1j(M1_1j*M2_1j)]

        Parameters
        ----------
        bulk_comp   dict
            Bulk composition of the major oxides in mol percent.

        b_number    int
            Number of B value to calculate.

        Returns
        -------
        float
            Partial B value for given b_number. E.g., if given b_number=1, will
            return value of B1. B, then, is the sum of all partial B's.
        """
        comp = bulk_comp.copy()

        # Convert FeO, Fe2O3 to FeOT
        comp['FeOT'] = comp['FeO']

        # map b numbers to oxides
        bs_to_oxides = {1: ['SiO2', 'TiO2'],
                        2: ['Al2O3'],
                        3: ['FeOT', 'MnO', 'P2O5'],
                        4: ['MgO'],
                        5: ['CaO'],
                        6: ['Na2O', 'H2O', 'F2O']}

        # known constant values (Table 1, Giordano et al., 2008)
        b_number_to_constants = {1: 159.56,
                                 2: -173.34,
                                 3: 72.13,
                                 4: 75.69,
                                 5: -38.98,
                                 6: -84.08,
                                 7: 141.54,
                                 11: -2.43,
                                 12: -0.91,
                                 13: 17.62}

        if b_number <= 6:
            partial_B_value = (b_number_to_constants[b_number] *
                               sum([comp[bs_to_oxides[b_number][x]] for x in
                                    range(len(bs_to_oxides[b_number]))]))
        if b_number == 7:
            partial_B_value = (b_number_to_constants[b_number] *
                               (comp['H2O'] + comp['F2O'] +
                               np.log(1+comp['H2O'])))

        if b_number == 11:
            partial_B_value = (b_number_to_constants[b_number] *
                               (comp['SiO2']+comp['TiO2']) *
                               (comp['FeOT']+comp['MnO']+comp['MgO']))

        if b_number == 12:
            partial_B_value = (b_number_to_constants[b_number] *
                               (comp['SiO2']+comp['TiO2']+comp['Al2O3'] +
                               comp['P2O5']) *
                               (comp['Na2O']+comp['K2O']+comp['H2O']))

        if b_number == 13:
            partial_B_value = (b_number_to_constants[b_number] *
                               (comp['Al2O3']) *
                               (comp['Na2O']+comp['K2O']))

        return partial_B_value

    def _calculate_partial_C_value(bulk_comp, c_number):
        """
        Calculate values for C (computed values in Giordano
        spreadsheet).

        C = sum(i=1-6)[c_i*N_i] + [c_11*(N1_11*N2_11)]

        Parameters
        ----------
        bulk_comp   dict
            Bulk composition of the major oxides in mol percent.

        c_number    int
            Number of C value to calculate.

        Returns
        -------
        float
            Partial C value for given c_number. E.g., if given c_number=1, will
            return value of C1. C, then, is defined by the equation above.
        """
        comp = bulk_comp.copy()

        # Convert FeO, Fe2O3 to FeOT
        comp['FeOT'] = comp['FeO']

        # map b numbers to oxides
        cs_to_oxides = {1: ['SiO2'],
                        2: ['TiO2', 'Al2O3'],
                        3: ['FeOT', 'MnO', 'MgO'],
                        4: ['CaO'],
                        5: ['Na2O', 'K2O']}

        # known constant values (Table 1, Giordano et al., 2008)
        c_number_to_constants = {1: 2.75,
                                 2: 15.72,
                                 3: 8.32,
                                 4: 10.20,
                                 5: -12.29,
                                 6: -99.54,
                                 11: 0.30}

        if c_number <= 5:
            partial_C_value = (c_number_to_constants[c_number] *
                               sum([comp[cs_to_oxides[c_number][x]] for x in
                                    range(len(cs_to_oxides[c_number]))]))
        if c_number == 6:
            partial_C_value = (c_number_to_constants[c_number] *
                               (np.log(1+comp['H2O'] + comp['F2O'])))

        if c_number == 11:
            partial_C_value = (c_number_to_constants[c_number] *
                               (comp['Al2O3'] + comp['FeOT'] + comp['MnO'] +
                                comp['MgO'] + comp['CaO'] - comp['P2O5']) *
                               (comp['Na2O'] + comp['K2O'] + comp['H2O'] +
                                comp['F2O']))

        return partial_C_value

    b_numbers = [1, 2, 3, 4, 5, 6, 7, 11, 12, 13]
    B_value = sum([_calculate_partial_B_value(comp_molpercent, b) for b in b_numbers])

    c_numbers = [1, 2, 3, 4, 5, 6, 11]
    C_value = sum([_calculate_partial_C_value(comp_molpercent, c) for c in c_numbers])

    log_nu_Pas = A_value + B_value/((temperature+273.15)-C_value)

    return round(log_nu_Pas, 4)



import pandas as pd
import os
import sys
import warnings as w

from VESIcal import core
from VESIcal import sample_class

# Turn off chained assignment pandas warning
pd.options.mode.chained_assignment = None  # default='warn'


def rename_duplicates(df, suffix='-duplicate-'):
    appendents = (suffix +
                  df.groupby(level=0).cumcount().astype(str).replace('0', ''))
    appendents = appendents.replace(suffix, '')
    return df.set_index(df.index.astype(str) + appendents)


class status_bar(object):
    """Various styles of status bars that display the progress of a calculation
    within a loop
    """
    def __init__():
        pass

    def status_bar(percent, sample_name=None, btext=None, barLen=20):
        """
        Prints an updating status bar to the terminal or jupyter notebook.

        Parameters
        ----------
        percent: float
            Percent value of progress from 0 to 1

        sample_name: string
            Name of the current sample being calculated

        btext: string
            Any extra text to display next to status bar

        barLen: int
            Length of bar to print
        """
        sys.stdout.write("\r")
        sys.stdout.write("[{:<{}}] {:.0f}%".format("=" * int(barLen * percent),
                                                   barLen, percent * 100))

        sample_string = str(sample_name)
        # Set max number of characters in sample name
        max_name_length = 25
        if len(str(sample_name)) >= max_name_length:
            sample_string = str(sample_name)[0:max_name_length-1] + "..."

        # Write out sample name and trailing spaces to cover contents of
        # previous sample names left over on line
        if sample_name is not None:
            sys.stdout.write("  Working on sample " + sample_string +
                             "                            ")
        if btext is not None:
            sys.stdout.write(" " + str(btext))
        if percent == 1.0:
            sys.stdout.write("\n")
        sys.stdout.flush()


# ---------- BATCHFILE CLASS --------- #
class BatchFile(object):
    """A batch file with sample names and oxide compositions

    Attributes
    ----------
        filename: str
            Path to the batch file, e.g., "my_file.xlsx". This always needs to
            be passed, even if the user is passing a pandas DataFrame rather
            than an batch file. If passing a DataFrame, filename should be set
            to None. File can be excel file (.xlsx) or .csv.

        sheet_name: str
            OPTIONAL. For Excel files. Default value is 0 which gets the first
            sheet in the batch spreadsheet file. This implements the pandas.
            read_excel() sheet_name parameter. But functionality to read in
            more than one sheet at a time (e.g.,
            pandas.read_excel(sheet_name=None)) is not yet imlpemented in
            VESIcal. From the pandas 1.0.4 documentation:

            Available cases:
            - Defaults to 0: 1st sheet as a DataFrame
            - 1: 2nd sheet as a DataFrame
            - "Sheet1": Load sheet with name “Sheet1”

        file_type: str
            OPTIONAL. Default is 'excel', which denotes that passed file has
            extension .xlsx. Other option is 'csv', which denotes that the
            passed file has extension .csv.

        units: str
            OPTIONAL. Default is 'wtpt_oxides'. String defining whether the
            oxide composition is given in wt percent ("wtpt_oxides", which is
            the default), mole oxides (mol_oxides) or mole cations
            (mol_cations).

        default_normalization:     None or str
            The type of normalization to apply to the data by default. One of:
            - None (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%,
               including volatiles. The volatile wt% will remain fixed, whilst
               the other major element oxides are reduced proportionally so
               that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to
              100%, assuming it is volatile-free. If H2O or CO2 are passed to
              the function, their un-normalized values will be retained in
              addition to the normalized non-volatile oxides, summing to >100%.

        default_units     str
            The type of composition to return by default, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations

        label: str
            OPTIONAL. Default is 'Label'. Name of the column within the passed
            file referring to sample names.

        dataframe: pandas DataFrame
            OPTIONAL. Default is None in which case this argument is ignored.
            This argument is used when the user wishes to turn a pandas
            DataFrame into an BatchFile object, for example when user data is
            already in python rather than being imported from a file. In this
            case set `dataframe` equal to the dataframe object being passed in.
            If using this option, pass None to filename.
    """
    def __init__(self, filename, sheet_name=0, file_type='excel',
                 units='wtpt_oxides', label='Label',
                 default_normalization='none', default_units='wtpt_oxides',
                 dataframe=None, **kwargs):
        """Return a BatchFile object whose parameters are defined here."""
        self.units = units
        self.set_default_normalization(default_normalization)
        self.set_default_units(default_units)

        if filename is not None:
            file_name, file_extension = os.path.splitext(filename)
            if file_extension == '.xlsx' or file_extension == '.xls':
                file_type = 'excel'
            if file_extension == '.csv':
                file_type = 'csv'

        if isinstance(sheet_name, str) or isinstance(sheet_name, int):
            pass
        else:
            raise core.InputError("If sheet_name is passed, it must be of "
                                  "type str or int. Currently, VESIcal cannot "
                                  "import more than one sheet at a time.")

        # handle data if passed in as existing dataframe or as file
        if dataframe is not None:
            data = dataframe
            if label is not None:
                data = self.try_set_index(data, label)
        else:
            if file_type == 'excel':
                data = pd.read_excel(filename, sheet_name=sheet_name)
                data = self.try_set_index(data, label)
            elif file_type == 'csv':
                data = pd.read_csv(filename)
                data = self.try_set_index(data, label)
            else:
                raise core.InputError("file_type must be one of \'excel\' or "
                                      "\'csv\'.")

        # Sanitize data inputs
        data = rename_duplicates(data)  # handle any duplicated sample names
        for column in data:  # convert all oxide columns to numeric
            if column in core.oxides:
                data[column] = data[column].apply(pd.to_numeric, errors='coerce')
        data = data.dropna(how='all')  # drop any rows that are all NaNs
        data = data.fillna(0)  # fill in any missing data with 0's

        if 'model' in kwargs:
            w.warn("You don't need to pass a model here, so it will be "
                   "ignored. You can specify a model when performing "
                   "calculations on your dataset (e.g., "
                   "calculate_dissolved_volatiles())",
                   RuntimeWarning, stacklevel=2)

        if 'norm' in kwargs:
            w.warn("We noticed you passed a norm argument here. This does "
                   "nothing. You can normalize your BatchFile and save it to "
                   "a new variable name after import using "
                   "normalize(BatchFileObject). See the documentation for "
                   "more info.",
                   RuntimeWarning, stacklevel=2)

        total_iron_columns = ["FeOt", "FeOT", "FeOtot", "FeOtotal", "FeOstar",
                              "FeO*"]
        for name in total_iron_columns:
            if name in data.columns:
                if 'FeO' in data.columns:
                    for row in data.itertuples():
                        if (data.at[row.Index, "FeO"] == 0 and
                           data.at[row.Index, name] > 0):
                            w.warn("Sample " + str(row.Index) + ": " +
                                   str(name) + " value of " +
                                   str(data.at[row.Index, name]) +
                                   " used as FeO. Fe2O3 set to 0.0.",
                                   RuntimeWarning, stacklevel=2)
                            data.at[row.Index, "Fe2O3"] = 0.0
                            data.at[row.Index, "FeO"] = (
                                                      data.at[row.Index, name])
                else:
                    w.warn("Total iron column " + str(name) + " detected. " +
                           "This column will be treated as FeO. If Fe2O3 " +
                           "data are not given, Fe2O3 will be 0.0. In " +
                           "future, an option to calcualte FeO/Fe2O3 based " +
                           "on fO2 will be implemented.",
                           RuntimeWarning, stacklevel=2)
                    data['FeO'] = data[name]

        if units == "wtpt_oxides":
            pass
        if units == "mol_oxides":
            data = self._molOxides_to_wtpercentOxides(data)
        if units == "mol_cations":
            data = self._molCations_to_wtpercentOxides(data)

        for column in data:
            if column in core.oxides:
                data[column][data[column] < 0] = 0

        self.data = data

    def set_default_normalization(self, default_normalization):
        """ Set the default type of normalization to use with the
        get_composition() method.

        Parameters
        ----------
        default_normalization:    str
            The type of normalization to apply to the data. One of:
            - 'none' (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%,
              including volatiles. The volatile wt% will remain fixed, whilst
              the other major element oxides are reduced proportionally so
              that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to
              100%, assuming it is volatile-free. If H2O or CO2 are passed to
              the function, their un-normalized values will be retained in
              addition to the normalized non-volatile oxides, summing to >100%.

        """
        if default_normalization in ['none', 'standard', 'fixedvolatiles',
                                     'additionalvolatiles']:
            self.default_normalization = default_normalization
        else:
            raise core.InputError("The normalization method must be one of "
                                  "'none', 'standard', 'fixedvolatiles' "
                                  "or 'additionalvolatiles'.")

    def set_default_units(self, default_units):
        """ Set the default units of composition to return when using the
        get_composition() method.

        Parameters
        ----------
        default_units     str
            The type of composition to return, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
        """
        if default_units in ['wtpt_oxides', 'mol_oxides', 'mol_cations']:
            self.default_units = default_units
        else:
            raise core.InputError("The units must be one of 'wtpt_oxides', "
                                  "'mol_oxides','mol_cations'.")

    def get_composition(self, species=None, normalization=None, units=None,
                        exclude_volatiles=False, asBatchFile=False):
        """ Returns a pandas DataFrame containing the compositional
        information for all samples in the BatchFile object

        Parameters
        ----------
        species:    NoneType or str
            The name of the oxide or cation to return the concentration of. If
            NoneType (default) the whole composition of each sample will be
            returned. If an oxide is passed, the value in wtpt will be
            returned unless units is set to 'mol_oxides', even if the default
            units for the sample object are mol_oxides. If an element is
            passed, the concentration will be returned as mol_cations, unless
            'mol_singleO' is specified as units, even if the default units for
            the sample object are mol_singleO. Unless normalization is
            specified in the method call, none will be applied.

        normalization:     NoneType or str
            The type of normalization to apply to the data. One of:
            - 'none' (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%,
              including volatiles. The volatile wt% will remain fixed, whilst
              the other major element oxides are reduced proportionally so
              that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to
              100%, assuming it is volatile-free. If H2O or CO2 are passed to
              the function, their un-normalized values will be retained in
              addition to the normalized non-volatile oxides, summing to >100%.
            If NoneType is passed the default normalization option will be
            used (self.default_normalization).

        units:     NoneType or str
            The units of composition to return, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
            - mol_singleO
            If NoneType is passed the default units option will be used
            (self.default_type).

        exclude_volatiles   bool
            If True, volatiles will be excluded from the returned composition,
            prior to normalization and conversion.

        asBatchFile:    bool
            If True, returns a BatchFile object. If False, returns a
            pandas.DataFrame object.

        Returns
        -------
        pandas.DataFrame or BatchFile object
            All sample information.
        """
        data = self.data.copy()

        # Fetch the default return types if not specified in function call
        if normalization is None and species is None:
            normalization = self.default_normalization
        if units is None and species is None:
            units = self.default_units

        new_compositions = []
        sample_names = []
        for index, row in data.iterrows():
            sample_comp = self.get_sample_composition(index, units=units,
                                                      asSampleClass=True)
            new_compositions.append(sample_comp.get_composition(
                     species=species, normalization=normalization, units=units,
                     exclude_volatiles=exclude_volatiles))
            sample_names.append(index)
        if isinstance(new_compositions[0], pd.Series):
            return_frame = pd.concat(
                           [pd.DataFrame(j) for j in new_compositions], axis=1)
            return_frame = return_frame.transpose()
            return_frame["new_index"] = sample_names
            return_frame = return_frame.set_index("new_index")
            return_frame.index.name = None
        elif isinstance(new_compositions[0], float):
            species_data = {species: new_compositions}
            return_frame = pd.DataFrame(
                           species_data, index=[name for name in sample_names])
        else:
            return_frame = None

        if asBatchFile is False:
            return return_frame
        else:
            return BatchFile(filename=None, dataframe=return_frame, label=None)

    def get_data(self, normalization=None, units=None, asBatchFile=False):
        """
        Returns all data stored in a BatchFile object (both compositional and
        other data). To return only the compositional data, use
        get_composition().

        Parameters
        ----------
        normalization:     NoneType or str
            The type of normalization to apply to the data. One of:
            - 'none' (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%,
              including volatiles. The volatile wt% will remain fixed, whilst
              the other major element oxides are reduced proportionally so
              that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to
              100%, assuming it is volatile-free. If H2O or CO2 are passed to
              the function, their un-normalized values will be retained in
              addition to the normalized non-volatile oxides, summing to >100%.

            If NoneType is passed the default normalization option will be
            used (self.default_normalization).

        units:     NoneType or str
            The units of composition to return, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
            - mol_singleO

            If NoneType is passed the default units option will be used
            (self.default_type).

        asBatchFile:    bool
            If True, returns a BatchFile object. If False, returns a
            pandas.DataFrame object.

        Returns
        -------
        pandas.DataFrame or BatchFile object
            All sample information.
        """
        data = self.data.copy()

        # Fetch the default return units if not specified in function call
        if units is None:
            units = self.default_units

        # Fetch the default normalization if not specified in the function call
        if normalization is None:
            normalization = self.default_normalization

        # Grab all compositional data
        compositional_data = self.get_composition(normalization=normalization,
                                                  units=units)

        # Grab all non-compositional data
        non_compositional_data = data.filter(
                       [col for col in data.columns if col not in core.oxides])

        # concatenate both compositional and non-compositional dataframes
        # into one
        return_frame = pd.concat([compositional_data, non_compositional_data],
                                 axis=1)

        if asBatchFile is False:
            return return_frame
        else:
            return BatchFile(filename=None, dataframe=return_frame, label=None)

    def get_sample_composition(self, samplename, species=None,
                               normalization=None, units=None,
                               asSampleClass=False):
        """
        Returns oxide composition of a single sample from a user-imported file
        as a dictionary

        Parameters
        ----------
        samplename: string
            Name of the desired sample

        normalization: NoneType or str
            The type of normalization to apply to the data. One of:
            - 'none' (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%,
              including volatiles. The volatile wt% will remain fixed, whilst
              the other major element oxides are reduced proportionally so
              that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to
              100%, assuming it is volatile-free. If H2O or CO2 are passed to
              the function, their un-normalized values will be retained in
              addition to the normalized non-volatile oxides, summing to >100%.

            If NoneType is passed the default normalization option will be
            used (self.default_normalization).

        units:     NoneType or str
            The units of composition to return, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
            - mol_singleO
            If NoneType is passed the default units option will be used
            (self.default_type).

        asSampleClass:  bool
            If True, the sample composition will be returned as a sample
            class, with default options. In this case any normalization
            instructions will be ignored.

        Returns
        -------
        dictionary, float, or sample_class.Sample object
            Composition of the sample as oxides
        """
        # Fetch the default return types if not specified in function call
        if normalization is None and species is None:
            normalization = self.default_normalization
        if units is None and species is None:
            units = self.default_units

        # Check that normalization being chosen is one of the possible options
        if normalization in [None, 'none', 'standard', 'fixedvolatiles',
                             'additionalvolatiles']:
            pass
        else:
            raise core.InputError("The normalization method must be one of "
                                  "'none', 'standard', 'fixedvolatiles', "
                                  "or 'additionalvolatiles'.")

        data = self.data
        my_sample = pd.DataFrame(data.loc[samplename])
        sample_dict = (my_sample.to_dict()[samplename])
        sample_oxides = {}
        for item, value in sample_dict.items():
            if item in core.oxides:
                sample_oxides.update({item: value})

        _sample = sample_class.Sample(sample_oxides)

        # Get sample composition in terms of any species, units, and
        # normalization passed
        return_sample = _sample.get_composition(species=species, units=units,
                                                normalization=normalization)

        if asSampleClass:
            return sample_class.Sample(return_sample)
        else:
            if species is None:
                return dict(return_sample)
            elif isinstance(species, str):
                return return_sample

    def _molOxides_to_wtpercentOxides(self, data):
        for i, row in data.iterrows():
            sample_comp = {}
            for oxide in core.oxides:
                if oxide in data.columns:
                    sample_comp[oxide] = row[oxide]
                else:
                    sample_comp[oxide] = 0.0
            _sample = sample_class.Sample(sample_comp, units='mol_oxides')
            _sample_conv = _sample.get_composition()
            for ox in core.oxides:
                data.loc[i, oxide] = _sample_conv[oxide]
        return data

    def _molCations_to_wtpercentOxides(self, data):
        for i, row in data.iterrows():
            sample_comp = {}
            for cation in core.oxides_to_cations[core.oxides]:
                if cation in data.columns:
                    sample_comp[cation] = row[cation]
                else:
                    sample_comp[cation] = 0.0
            _sample = sample_class.Sample(sample_comp, units='mol_cations')
            _sample_conv = _sample.get_composition()
            for oxide in core.oxides:
                data.loc[i, oxide] = _sample_conv[oxide]
        return data

    def try_set_index(self, dataframe, label):
        """
        Method to handle setting the index column in an BatchFile object. If
        no column is passed that matches the default index name, then this
        method will attempt to choose the 'best' column that the user might
        want to serve as an index column.

        Parameters
        ----------
        dataframe: pandas DataFrame

        label: str
            Name of the column within the passed Excel file referring to
            sample names.
        """
        _dataframe = dataframe.copy()
        try:
            _dataframe = _dataframe.set_index(label)
        except Exception:
            label_found = False
            for col in _dataframe.columns:
                if col in core.oxides:
                    pass
                else:
                    _dataframe = _dataframe.set_index(col)
                    label_found = True
                    w.warn("No Label column given, so column '" + str(col) +
                           "' was chosen for you. To choose your own, set " +
                           "label='<column-name>'.", RuntimeWarning,
                           stacklevel=2)
                    break
            if label_found is False:
                _dataframe.index.name = 'Label'
                w.warn("No Label column given, so one was created for you. "
                       "To choose your own, set label='<column-name>'.",
                       RuntimeWarning, stacklevel=2)

        return _dataframe

    def save_excel(self, filename, calculations, sheet_names=None):
        """
        Saves data calculated by the user in batch processing mode (using the
        BatchFile class methods) to an organized Excel file, with the original
        user data plus any calculated data.

        Parameters
        ----------
        filename: string
            Name of the file. Extension (.xlsx) should be passed along with
            the name itself, all in quotes (e.g., 'myfile.xlsx').

        calculations: pandas DataFrame or list of pandas DataFrames
            A single DataFrame or list of DataFrames (e.g., calculated outputs
            from any of the core BatchFile functions:
            calculate_dissolved_volatiles, calculate_equilibrium_fluid_comp,
            and calculate_saturation_pressure). If None, only the original
            user data will be saved.

        sheet_names: None, string, or list
            OPTIONAL. Default value is None. Allows user to set the name of
            the sheet or sheets written to the Excel file.

        Returns
        -------
            Creates and saves an Excel file with data from each calculation
            saved to its own sheet.
        """
        if isinstance(calculations, list):
            if isinstance(sheet_names, list) or sheet_names is None:
                pass
            else:
                raise core.InputError("If calculations is passed as list, "
                                      "sheet_names must also be list of same "
                                      "length")
        elif calculations is None:
            pass
        else:
            calculations = [calculations]

        with pd.ExcelWriter(filename) as writer:
            self.data.to_excel(writer, 'Original_User_Data')
            if isinstance(calculations, list):
                if sheet_names is None:
                    for n, df in enumerate(calculations):
                        df.to_excel(writer, 'Calc%s' % n)
                elif isinstance(sheet_names, list):
                    pass
                else:
                    sheet_names = [sheet_names]
                if isinstance(sheet_names, list):
                    if len(sheet_names) == len(calculations):
                        pass
                    else:
                        raise core.InputError("calculations and sheet_names "
                                              "must have the same length")

                    for i in range(len(calculations)):
                        if isinstance(sheet_names[i], str):
                            calculations[i].to_excel(writer, sheet_names[i])
                        else:
                            raise core.InputError("if sheet_names is passed, "
                                                  "it must be list of strings")
            elif calculations is None:
                pass
        return print("Saved " + str(filename))

    def save_csv(self, filenames, calculations, **kwargs):
        """
        Saves data calculated by the user in batch processing mode to a
        comma-separated values (csv) file. Mirros the pandas.to_csv() method.
        Any argument that can be passed to pandas.csv() can be passed here.
        One csv file will be saved for each calculation passed.

        Parameters
        ----------
        filenames: string or list of strings
            Name of the file. Extension (.csv) should be passed along with
            the name itself, all in quotes (e.g., 'myfile.csv'). The number
            of calculations passed must match the number of filenames passed.
            If passing more than one, should be passed as a list.

        calculations: pandas DataFrame or list of pandas DataFrames
            A single variable or list of variables containing calculated
            outputs from any of the core BatchFile functions:
            calculate_dissolved_volatiles, calculate_equilibrium_fluid_comp,
            and calculate_saturation_pressure.

        Returns
        -------
            Creates and saves a CSV file or files with data from each
            calculation saved to its own file.
        """
        if type(filenames) != list:
            filenames = [filenames]
        if type(calculations) != list:
            calculations = [calculations]
        if len(filenames) != len(calculations):
            raise core.InputError("calculations and filenames must have the "
                                  "same length")

        for i in range(len(filenames)):
            calculations[i].to_csv(filenames[i], **kwargs)
            print("Saved " + str(filenames[i]))


def from_DataFrame(dataframe, units='wtpt_oxides', label='Label'):
    """
    Transforms any pandas DataFrame object into a VESIcal BatchFile object.

    Parameters
    ----------
    dataframe: pd.DataFrame object
        DataFrame object containing samples and oxide compositions.

    units: str
        OPTIONAL. Default is 'wtpt_oxides'. String defining whether the oxide
        composition is given in wt percent ("wtpt_oxides", which is the
        default), mole fraction oxides ("mol_oxides"), or mole fraction
        cations ("mol_cations").

    label: str
        OPTIONAL. Default is 'Label'. Name of the column within the passed
        file referring to sample names. This column will be set as the index
        column.

    Returns
    -------
    VESIcal.BatchFile object
    """
    return BatchFile(filename=None, dataframe=dataframe, units=units,
                     label=label)



from abc import abstractmethod
import warnings as w
import numpy as np

from VESIcal import core
from VESIcal import models
from VESIcal.models import magmasat

from copy import deepcopy


class Calculate(object):
    """ The Calculate object is a template for implementing user-friendly
    methods for running calculations using the volatile solubility models.
    All Calculate methods have a common workflow- sample is read in,
    preprocessed, the calculation is performed, the calibration range is
    checked, and the results stored.
    """
    def __init__(self, sample, model='MagmaSat', silence_warnings=False,
                 **kwargs):
        """
        Initializes the calculation.

        Parameters
        ----------
        sample:    Sample class
            The rock composition as a Sample object.
        model:     string or Model class
            Which model to use for the calculation. If passed a string, it
            will look up the name in the default_models dictionary. Default is
            MagmaSat.
        silence_warnings:     bool
            Silence warnings about calibration ranges. Default is False.
        preprocess_sample:     bool
            Before running the calculation, run the sample through the
            preprocessing routine. As of Feb 2021 this functionality should be
            redundant.
        """
        self.model_name = model
        if model == 'MagmaSat':
            self.model = magmasat.MagmaSat()
        elif type(model) == str:
            if model in models.default_models.keys():
                self.model = models.default_models[model]
            else:
                raise core.InputError("The model name given is not recognised."
                                      " Run the method get_model_names() to "
                                      "find allowed names.")
        else:
            self.model = model

        self.sample = sample

        self.result = self.calculate(sample=self.sample, **kwargs)
        self.calib_check = self.check_calibration_range(sample=self.sample,
                                                        **kwargs)

        if self.calib_check is not None and silence_warnings is False:
            if self.calib_check != '':
                w.warn(self.calib_check, RuntimeWarning)

    @abstractmethod
    def calculate(self):
        """ """

    @abstractmethod
    def check_calibration_range(self):
        """ """


class calculate_dissolved_volatiles(Calculate):
    """ Calculates the dissolved volatile concentration using a chosen model
    (default is MagmaSat). Using this interface will preprocess the sample,
    run the calculation, and then check the calibration ranges. All parameters
    required by the chosen model must be passed.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    pressure:   float
        Total pressure in bars.
    model:  string or Model object
        Model to be used. If using one of the default models, this can be
        the string corresponding to the model in the default_models dict.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when
        calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the
        preprocessing operations within the models. If you obtain unexpected
        results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access results by fetching the result property.
        Dissolved volatile concentrations (in wt%), in order (CO2, H2O, if
        using a mixed fluid default model).
    """
    def return_default_units(self, sample, calc_result, **kwargs):
        """ Checkes the default units set for the sample_class.Sample
        object and returns the result of a calculation in those units.

        Parameters
        ----------
        sample: Sample class
            The rock composition as a Sample object.

        calc_result: dict or float
            Result of a calculate_dissolved_volatiles() calculation on a
            sample.
        """
        default_units = sample.default_units

        # get the composition of
        bulk_comp = deepcopy(sample)

        # check if calculation result is H2O-only, CO2-only, or mixed
        # H2O-CO2
        if isinstance(self.model_name, str):
            if self.model_name in models.get_model_names(model='mixed'):
                # set dissolved H2O and CO2 values.
                # this action assumes they are input as wt% but updates the
                # composition in its default units.
                bulk_comp.change_composition({'H2O': calc_result['H2O_liq'],
                                             'CO2': calc_result['CO2_liq']})
                return {'H2O_liq': bulk_comp.get_composition(
                                                      species='H2O',
                                                      units=default_units),
                        'CO2_liq': bulk_comp.get_composition(
                                                      species='CO2',
                                                      units=default_units)}
            elif 'Water' in self.model_name:
                bulk_comp.change_composition({'H2O': calc_result})
                return bulk_comp.get_composition(species='H2O',
                                                 units=default_units)
            elif 'Carbon' in self.model_name:
                bulk_comp.change_composition({'CO2': calc_result})
                return bulk_comp.get_composition(species='CO2',
                                                 units=default_units)
            elif self.model_name == 'MagmaSat':
                bulk_comp.change_composition({'H2O': calc_result['H2O_liq'],
                                             'CO2': calc_result['CO2_liq']})
                # check if verbose method has been chosen
                if 'verbose' in kwargs and kwargs['verbose']:
                    return {'H2O_liq': bulk_comp.get_composition(
                                                      species='H2O',
                                                      units=default_units),
                            'CO2_liq': bulk_comp.get_composition(
                                                      species='CO2',
                                                      units=default_units),
                            'XH2O_fl': calc_result['XH2O_fl'],
                            'XCO2_fl': calc_result['XCO2_fl'],
                            'FluidProportion_wt': calc_result[
                                                     'FluidProportion_wt']}
                else:
                    return {'H2O_liq': bulk_comp.get_composition(
                                                      species='H2O',
                                                      units=default_units),
                            'CO2_liq': bulk_comp.get_composition(
                                                      species='CO2',
                                                      units=default_units)}
        else:
            # if self.model_name is not a string, most likely this is a
            # user-created model class, and so we cannot interrogate it for
            # model type (H2O, CO2, or mixed).
            # TODO: create model type parameter so that we can ID model type
            # this way instead of via strings in a list!
            return calc_result

    def calculate(self, sample, pressure, **kwargs):
        dissolved = self.model.calculate_dissolved_volatiles(pressure=pressure,
                                                             sample=sample,
                                                             returndict=True,
                                                             **kwargs)
        dissolved_default_units = self.return_default_units(sample, dissolved,
                                                            **kwargs)
        return dissolved_default_units

    def check_calibration_range(self, sample, pressure, **kwargs):
        parameters = kwargs
        parameters['sample'] = sample
        parameters.update(dict(sample.get_composition(units='wtpt_oxides',
                                                      normalization='none')))
        parameters['pressure'] = pressure
        if len(self.model.volatile_species) == 1:
            volspec = self.model.volatile_species[0]
            volconc = self.result
            parameters.update({volspec: volconc})
        else:
            parameters.update(self.result)

        calib_check = self.model.check_calibration_range(parameters)
        return calib_check


class calculate_equilibrium_fluid_comp(Calculate):
    """ Calculates the equilibrium fluid composition using a chosen model
    (default is MagmaSat). Using this interface will preprocess the sample,
    run the calculation, and then check the calibration ranges. All parameters
    required by the chosen model must be passed.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    pressure:   float or None
        Total pressure in bars. If None, the saturation pressure will be used.
    model:  string or Model object
        Model to be used. If using one of the default models, this can be
        the string corresponding to the model in the default_models dict.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when
        calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the
        preprocessing operations within the models. If you obtain unexpected
        results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access result by fetching the result property. Mole
        fractions of each volatile species, in order (CO2, then H2O, if using
        a mixed-fluid default model).
    """
    def calculate(self, sample, pressure=None, **kwargs):
        if pressure is None:
            pressure = float(self.model.calculate_saturation_pressure(
                                                                 sample=sample,
                                                                 verbose=False,
                                                                 **kwargs))

        fluid_comp = self.model.calculate_equilibrium_fluid_comp(
                                                             pressure=pressure,
                                                             sample=sample,
                                                             **kwargs)
        return fluid_comp

    def check_calibration_range(self, sample, pressure=None, **kwargs):
        if pressure is None:
            pressure = self.model.calculate_saturation_pressure(sample=sample,
                                                                **kwargs)
        parameters = kwargs
        parameters.update(dict(sample.get_composition(units='wtpt_oxides',
                                                      normalization='none')))
        parameters['sample'] = sample
        parameters['pressure'] = pressure
        if len(self.model.volatile_species) == 1:
            volspec = self.model.volatile_species
            volconc = {volspec[0]: self.result}
            parameters.update(volconc)
        elif type(self.model.volatile_species) == list:
            parameters.update(self.result)

        calib_check = self.model.check_calibration_range(parameters)
        return calib_check


class calculate_isobars_and_isopleths(Calculate):
    """ Calculates isobars and isopleths using a chosen model (default is
    MagmaSat). Using this interface will preprocess the sample, run the
    calculation, and then check the calibration ranges. All parameters
    required by the chosen model must be passed.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    pressure_list:   list
        List of all pressure values at which to calculate isobars, in bars.
    isopleth_list:   list
        OPTIONAL: Default value is None, in which case only isobars will be
        calculated. List of all fluid compositions in mole fraction (of the
        first species in self.volatile_species) at which to calcualte
        isopleths. Values can range from 0 to 1.
    points:     int
        The number of points in each isobar and isopleth. Default value is 101.
    model:  string or Model object
        Model to be used. If using one of the default models, this can be
        the string corresponding to the model in the default_models dict.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when
        calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the
        preprocessing operations within the models. If you obtain unexpected
        results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access results by fetching the result property.
        If isopleth_list is not None, two objects will be returned, one with
        the isobars and the second with the isopleths. If return_dfs is True,
        two pandas DataFrames will be returned with column names
        'Pressure' or 'XH2O_fl', 'H2O_liq', and 'CO2_liq'. If return_dfs is
        False, two lists of numpy arrays will be returned. Each array is an
        individual isobar or isopleth, in the order passed via pressure_list
        or isopleth_list. The arrays are the concentrations of H2O and CO2 in
        the liquid, in the order of the species in self.volatile_species.
    """
    def calculate(self, sample, pressure_list, isopleth_list=[0, 1],
                  points=101, **kwargs):
        check = getattr(self.model, "calculate_isobars_and_isopleths", None)
        if callable(check):
            # samplenorm = sample.copy()
            # samplenorm = normalize_AdditionalVolatiles(samplenorm)
            isobars, isopleths = self.model.calculate_isobars_and_isopleths(
                               sample=self.sample, pressure_list=pressure_list,
                               isopleth_list=isopleth_list, points=points,
                               **kwargs)
            return isobars, isopleths
        else:
            raise core.InputError("This model does not have a "
                                  "calculate_isobars_and_isopleths method "
                                  "built in, most likely because it is a pure "
                                  "fluid model.")

    def check_calibration_range(self, sample, pressure_list, **kwargs):
        parameters = kwargs
        parameters.update(dict(sample.get_composition(units='wtpt_oxides')))
        parameters['sample'] = sample
        s = ''
        s += self.model.check_calibration_range(parameters)
        parameters = {}
        if isinstance(pressure_list, list):
            pass
        else:
            pressure_list = [pressure_list]
        for pressure in pressure_list:
            parameters['pressure'] = pressure
            s += self.model.check_calibration_range(parameters,
                                                    report_nonexistance=False)
        return s


class calculate_saturation_pressure(Calculate):
    """
    Calculates the pressure at which a fluid will be saturated, given the
    dissolved volatile concentrations. Using this interface will preprocess
    the sample, run the calculation, and then check the calibration ranges.
    All parameters required by the chosen model must be passed.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    model:  string or Model object
        Model to be used. If using one of the default models, this can be
        the string corresponding to the model in the default_models dict.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when
        calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the
        preprocessing operations within the models. If you obtain unexpected
        results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access results by fetching the result property.
        The saturation pressure in bars as a float.
    """
    def calculate(self, sample, **kwargs):
        satP = self.model.calculate_saturation_pressure(sample=sample,
                                                        **kwargs)
        return satP

    def check_calibration_range(self, sample, **kwargs):
        parameters = kwargs
        if isinstance(self.result, dict):  # handles cases where verbose=True
            parameters['pressure'] = next(iter(self.result.values()))
        else:
            parameters['pressure'] = self.result
        parameters.update(dict(sample.get_composition(units='wtpt_oxides')))
        parameters['sample'] = sample
        s = self.model.check_calibration_range(parameters)
        return s


class calculate_degassing_path(Calculate):
    """
    Calculates the dissolved volatiles in a progressively degassing sample.

    Parameters
    ----------
    sample:    Sample class
        The rock composition as a Sample object.
    pressure     string, float, int, list, or numpy array
        Defaults to 'saturation', the calculation will begin at the saturation
        pressure. If a number is passed as either a float or int, this will be
        the starting pressure. If a list of numpy array is passed, the
        pressure values in the list or array will define the degassing path,
        i.e. final_pressure and steps variables will be ignored. Units are
        bars.
    fractionate_vapor     float
        What proportion of vapor should be removed at each step. If 0.0
        (default), the degassing path will correspond to closed-system
        degassing. If 1.0, the degassing path will correspond to open-system
        degassing.
    final_pressure         float
        The final pressure on the degassing path, in bars. Ignored if a list
        or numpy array is passed as the pressure variable. Default is 1 bar.
    steps     int
        The number of steps in the degassing path. Ignored if a list or
        numpy array are passed as the pressure variable.
    model:  string or Model object
        Model to be used. If using one of the default models, this can be
        the string corresponding to the model in the default_models dict.
    silence_warnings     bool
        If set to True, no warnings will be raised automatically when
        calibration checks fail.
    preprocess_sample     bool
        If True (default), the sample will be preprocessed according to the
        preprocessing operations within the models. If you obtain unexpected
        results, try setting to False.

    Returns
    -------
    Calculate object
        Calculate object, access results by fetching the result property.
        A DataFrame with columns 'Pressure', 'H2O_liq', 'CO2_liq',
        'H2O_fl', 'CO2_fl', and 'FluidProportion_wt', is returned. Dissolved
        volatiles are in wt%, the proportions of volatiles in the fluid are
        in mole fraction.
    """
    def calculate(self, sample, pressure='saturation', fractionate_vapor=0.0,
                  final_pressure=100.0, **kwargs):
        check = getattr(self.model, "calculate_degassing_path", None)
        if callable(check):
            data = self.model.calculate_degassing_path(
                                           sample=sample, pressure=pressure,
                                           fractionate_vapor=fractionate_vapor,
                                           final_pressure=final_pressure,
                                           **kwargs)
            return data
        else:
            raise core.InputError("This model does not have a "
                                  "calculate_isobars_and_isopleths method "
                                  "built in, most likely because it is a pure "
                                  "fluid model.")

    def check_calibration_range(self, sample, **kwargs):
        parameters = kwargs
        parameters.update(dict(sample.get_composition()))
        parameters['sample'] = sample
        s = self.model.check_calibration_range(parameters)
        parameters = {}
        parameters['pressure'] = np.nanmax(self.result.Pressure_bars)
        s += self.model.check_calibration_range(parameters,
                                                report_nonexistance=False)
        return s



from copy import copy
import numpy as np


class CalibrationRange(object):
    """ The CalibrationRange object allows the range of allowable parameters
    to be specified and used in checking and reporting of the results.
    """
    def __init__(self, parameter_name, value, checkfunction=None, units='',
                 model_name='', fail_msg='', fail_dict={}, pass_msg='',
                 pass_dict={}, description_msg='', description_dict={}):
        self.parameter_name = parameter_name
        self.value = value
        self.checkfunction = checkfunction
        self.units = units
        self.model_name = model_name
        self.fail_msg = (copy(fail_msg), copy(fail_dict))
        self.pass_msg = (copy(pass_msg), copy(pass_dict))
        self.description_msg = (copy(description_msg), copy(description_dict))

    def check(self, parameters):
        """Method for checking whether parameters satisfy the calibration
        range.
        """
        if self.parameter_name in parameters:
            return self.checkfunction(self.value,
                                      parameters[self.parameter_name])
        else:
            return None

    def string(self, parameters, report_nonexistance=True):
        """Returns a string statement of the calibration check"""
        if parameters is None:
            msgdict = self.description_msg[1]
            if type(self.value) == float or type(self.value) == int:
                msgdict['calib_val'] = self.value
            elif (type(self.value) == list or type(self.value) == tuple or
                  type(self.value) == np.ndarray):
                for i in range(len(self.value)):
                    msgdict['calib_val'+str(i)] = self.value[i]
            if 'param_name' not in msgdict:
                msgdict['param_name'] = self.parameter_name
            if 'units' not in msgdict:
                msgdict['units'] = self.units
            if 'model_name' not in msgdict:
                msgdict['model_name'] = self.model_name
            return self.description_msg[0].format(**msgdict)
        else:
            check = self.check(parameters)
            if check:
                msgdict = self.pass_msg[1]
                msgdict['param_val'] = parameters[self.parameter_name]
                if type(self.value) == float or type(self.value) == int:
                    msgdict['calib_val'] = self.value
                elif (type(self.value) == list or type(self.value) == tuple or
                      type(self.value) == np.ndarray):
                    for i in range(len(self.value)):
                        msgdict['calib_val'+str(i)] = self.value[i]
                if 'param_name' not in msgdict:
                    msgdict['param_name'] = self.parameter_name
                if 'units' not in msgdict:
                    msgdict['units'] = self.units
                if 'model_name' not in msgdict:
                    msgdict['model_name'] = self.model_name
                return self.pass_msg[0].format(**msgdict)
            elif check is False:
                msgdict = self.fail_msg[1]
                msgdict['param_val'] = parameters[self.parameter_name]
                if type(self.value) == float or type(self.value) == int:
                    msgdict['calib_val'] = self.value
                elif (type(self.value) == list or type(self.value) == tuple or
                      type(self.value) == np.ndarray):
                    for i in range(len(self.value)):
                        msgdict['calib_val'+str(i)] = self.value[i]
                if 'param_name' not in msgdict:
                    msgdict['param_name'] = self.parameter_name
                if 'units' not in msgdict:
                    msgdict['units'] = self.units
                if 'model_name' not in msgdict:
                    msgdict['model_name'] = self.model_name
                return self.fail_msg[0].format(**msgdict)
            else:
                if report_nonexistance:
                    return "A value for {} was not provided.".format(
                                                           self.parameter_name)
                else:
                    return ''

# ------------- DEFAULT CALIBRATIONRANGE OBJECTS --------------- #


def crf_EqualTo(calibval, paramval):
    return calibval == paramval


crmsg_EqualTo_pass = ("The {param_name} ({param_val:.1f} {units}) is equal to "
                      "{calib_val:.1f} {units} as required by the calibration "
                      "range of the {model_name} model. ")
crmsg_EqualTo_fail = ("{param_name} ({param_val:.1f} {units}) is outside the "
                      "calibration range of the {model_name} model "
                      "({calib_val:.1f} {units}). ")
crmsg_EqualTo_description = ("The {model_name} model is calibrated for "
                             "{param_name} equal to {calib_val:.1f} {units}. ")


def crf_MixedFluidWarning(calibval, paramval):
    if type(paramval) == float or type(paramval) == int or len(paramval) == 1:
        return paramval == 0 or paramval == 1
    elif len(paramval) == 2:
        return list(paramval) == [0, 1] or list(paramval) == [1, 0]
    else:
        return True


crmsg_MixedFluidWarning_pass = "The pure fluid model is being used. "
crmsg_MixedFluidWarning_fail = ("The {model_name} model should be used to "
                                "model mixed fluids with caution. The mixed "
                                "fluid model does not recreate the "
                                "experimental observations. ")
crmsg_MixedFluidWarning_description = ("The {model_name} shoule be used to "
                                       "model mixed fluids with caution. ")


def crf_GreaterThan(calibval, paramval):
    return paramval >= calibval


crmsg_GreaterThan_pass = ("The {param_name} ({param_val:.1f} {units}) is "
                          "greater than {calib_val:.1f} {units} as required "
                          "by the calibration range of the {model_name} "
                          "model. ")
crmsg_GreaterThan_fail = ("The {param_name} is outside the calibration range "
                          "of {model_name} ({param_val:.1f}<{calib_val:.1f} "
                          "{units}. ")
crmsg_GreaterThan_description = ("The {model_name} model is calibrated for "
                                 "{param_name} greater than {calib_val:.1f} "
                                 "{units}. ")


def crf_LessThan(calibval, paramval):
    return paramval <= calibval


crmsg_LessThan_pass = ("The {param_name} ({param_val:.1f} {units}) is less "
                       "than {calib_val:.1f} {units} as required by the "
                       "calibration range of the {model_name} model. ")
crmsg_LessThan_fail = ("The {param_name} is outside the calibration range "
                       "of {model_name} ({param_val:.1f}>{calib_val:.1f} "
                       "{units}. ")
crmsg_LessThan_description = ("The {model_name} model is calibrated for "
                              "{param_name} less than {calib_val:.1f} "
                              "{units}. ")


def crf_Between(calibval, paramval):
    return paramval >= calibval[0] and paramval <= calibval[1]


crmsg_Between_pass = ("The {param_name} ({param_val:.1f} {units}) is between "
                      "{calib_val0:.1f} and {calib_val1:.1f} {units} as "
                      "required by the calibration range of the {model_name} "
                      "model. ")
crmsg_Between_fail = ("{param_name} ({param_val:.1f} {units}) is outside the "
                      "calibration range of the {model_name} model "
                      "({calib_val0:.1f}-{calib_val1:.1f} {units}). ")
crmsg_Between_description = ("The {model_name} model is calibrated for "
                             "{param_name} between {calib_val0:.1f} and "
                             "{calib_val1:.1f} {units}. ")
# Different wording for compositional checks (loosing "the")
crmsg_BC_pass = ("{param_name} ({param_val:.1f} {units}) is between "
                 "{calib_val0:.1f} and {calib_val1:.1f} {units} as required "
                 "by the calibration range of the {model_name} model. ")
crmsg_BC_fail = ("{param_name} ({param_val:.1f} {units}) is outside the "
                 "calibration range ({calib_val0:.1f}-{calib_val1:.1f} "
                 "{units}). ")
